import copy
import datetime
import os
import random

from yangke.performance.iapws97 import get_h_by_pt
from yangke.common.config import logger
from yangke.common.fileOperate import write_excel, pd_set_na, read_excel
from yangke.sis.BeiJingShenHua.tags import TagsRead, TagsWrite
from yangke.sis.dll_file import read_data, get_tag_value, init_write_sis, init_dbp_api
import pandas as pd
from yangke.sis.BeiJingShenHua.tags import get_pump_type_by_int
from yangke.base import is_number

ip = "10.2.128.168"
port = "12084"
user = "admin"
passwd = "admin"

fit_index = {
    "A3": [],
    "A4": [],
    "A5": [],
    "A6": [],
    "A7": [],
    "A8": [],
    "B3": [],
    "B4": [],
    "B5": [],
    "B6": [],
    "B7": [],
    "B8": [],
    "C3": [],
    "C4": [],
    "C5": [],
    "C6": [],
    "C7": [],
    "C8": [],
    "D3": [-0.006961224, 0.010512006, 2.200054686],
    "D4": [],
    "D5": [],
    "D6": [],
    "D7": [],
    "D8": [],
    "E3": [],
    "E4": [],
    "E5": [],
    "E6": [],
    "E7": [],
    "E8": [],
    "F3": [],
    "F4": [],
    "F5": [],
    "F6": [],
    "F7": [],
    "F8": [],
    "G3": [],
    "G4": [],
    "G5": [],
    "G6": [],
    "G7": [],
    "G8": [],
    "H3": [],
    "H4": [],
    "H5": [],
    "H6": [],
    "H7": [],
    "H8": [],
    "I3": [],
    "I4": [],
    "I5": [],
    "I6": [],
    "I7": [],
    "I8": [],
}


def filter_(df, tag, value_range):
    """
    从df中删除参数不在范围内的数据，示例
    filter_(his_value, "#1电功率", [200,300])  # 只保留电功率在200~300之间的行
    filter_(his_value, "#1电功率", [200,None]) # 只保留电功率在200以上的行
    filter_(his_value, "#1电功率", [None,100])  # 只保留电功率在100以下的行
    """
    min_value = value_range[0]
    max_value = value_range[1]
    if min_value is not None:
        df = df[df[tag] >= min_value]
    if max_value is not None:
        df = df[df[tag] <= max_value]
    return df


def get_history(tags, des, months=12, days=0, start_time: datetime.datetime = None, end_time: datetime.datetime = None,
                time_interval=60, save_to: str | None = None):
    """
    从RDBProxy代理服务器上导出历史数据，如果导出数据的时标是1970年，请检查tags标签是否正确。
    本方法要加载settings.yaml文件中的RDBP服务器信息。
    默认下载最近一年的历史数据。

    :param tags: list类型，导出测点的标签，需要与DCS系统上测点的KKS码或标签名一致
    :param des: list类型，导出测点的描述，可以自定义，用于注释导出的测点，如果传入该参数，则长度必须与tags相同
    :param months: 如果需要下载最近n月的历史数据，通过该参数传入下载历史数据的月数，如果按起止时间下载历史数据，无需设置该参数
    :param days: 如果需要下载最近n天的历史数据，通过该参数传入下载历史数据的天数，如果按起止时间下载历史数据，无需设置该参数
    :param start_time: 下载历史数据的起始时间，如果下载最近n月/天的历史数据，该参数留空
    :param end_time: 下载历史数据的截止时间，如果下载最近n月/天的历史数据，该参数留空
    :param time_interval: 下载数据的时间间隔，单位为秒，默认60s
    :param save_to: 是否保存到文件，默认不保存，如果需要保存，传入文件名
    """
    from yangke.base import merge_two_dataframes

    if start_time is not None and end_time is not None:  # 按起止时间下载历史数据
        if start_time >= end_time:
            return
        now = datetime.datetime.now()
        if start_time > now:
            return
        if end_time > now:
            end_time = now

        his_values = []
        delta_time = datetime.timedelta(days=1)
        _end = start_time + delta_time
        while _end < end_time:
            api = init_dbp_api()
            his_value = api.get_his_value(tags=tags, tags_description=des, start_time=start_time, end_time=_end,
                                          time_interval=time_interval)
            his_values.append(his_value)
            _end = _end + delta_time
            api.close()
        api = init_dbp_api()
        his_value = api.get_his_value(tags=tags, tags_description=des, start_time=_end, end_time=end_time,
                                      time_interval=time_interval)
        api.close()
        his_values.append(his_value)
        res = his_values[0]
        for v in his_values[1:]:
            res = merge_two_dataframes(res, v)
    else:
        now = datetime.datetime.now()
        if months != 0:
            _start_time = now - datetime.timedelta(days=months * 30, hours=0)
            _end_time = now
            res = get_history(tags, des, start_time=_start_time, end_time=_end_time, time_interval=time_interval,
                              save_to=None)
        else:
            # 下载最后days天的历史数据
            _start_time = now - datetime.timedelta(days=days, hours=0)
            _end_time = now
            res = get_history(tags, des, start_time=_start_time, end_time=_end_time, time_interval=time_interval,
                              save_to=None)
    if save_to:
        res.to_csv(save_to)
    return res


def deal_data(files=None, work_folder=None):
    if files is None:
        files = ["origin_1.csv", "origin_2.csv", "origin_3.csv", "origin_4.csv",
                 "origin_5.csv", "origin_6.csv", "origin_7.csv", "origin_8.csv",
                 "origin_9.csv", "origin_11.csv",
                 ]

    output_data = {}
    for file in files:
        logger.debug(f"处理{file=}")
        if work_folder is not None:
            file = os.path.join(work_folder, file)
        df = pd.read_csv(file)
        his = df.copy()
        df.set_index("DateTime", inplace=True)
        df.index = pd.DatetimeIndex(df.index)  # 将索引类型转换为DatetimeIndex以便可以使用按时间的插值方法
        # df = df.iloc[0:40]
        df["大气压力"] = df["大气压力"] / 10000  # MPa
        # 将大气压力低于0.03MPa的行的数据全部置为np.nan，因为发现有时候DCS系统数据传输故障，时序数据库中数据全都为0
        pd_set_na(df, {"大气压力": {"<": 0.03}}, expand=True, axis=0)
        df = df.interpolate(method="time")  # 31332
        df["天然气流量"] = df["计量单元天然气流量1"] + df["计量单元天然气流量2"] + df["计量单元天然气流量3"]
        df.drop(columns=["建议循泵运行方式", "建议风机运行方式", "最优运行方式下背压", "优化后机组节煤量",
                         "计量单元天然气流量1", "计量单元天然气流量2", "计量单元天然气流量3",
                         ], axis=1, inplace=True)
        for _fan_num in ["风机1电流", "风机2电流", "风机3电流", "风机4电流", "风机5电流", "风机6电流", "风机7电流",
                         "风机8电流",
                         "大泵1电流", "大泵2电流", "小高速3电流", "小高速4电流", "小低速3电流", "小低速4电流"]:
            df[_fan_num] = df[_fan_num].apply(lambda x: 0 if x <= 10 else x)
            df[_fan_num] = df[_fan_num].apply(lambda x: 1 if x > 10 else x)
        df["机力塔数量"] = df["风机1电流"] + df["风机2电流"] + df["风机3电流"] + df["风机4电流"] + \
                           df["风机5电流"] + df["风机6电流"] + df["风机7电流"] + df["风机8电流"]
        df["机力塔数量"] = df["机力塔数量"].apply(lambda x: int(x))
        df["当前大泵数量"] = df["大泵1电流"] + df["大泵2电流"]
        df["当前高速泵数量"] = df["小高速3电流"] + df["小高速4电流"]
        df["当前低速泵数量"] = df["小低速3电流"] + df["小低速4电流"]
        df.drop(columns=["风机1电流", "风机2电流", "风机3电流", "风机4电流", "风机5电流", "风机6电流", "风机7电流",
                         "风机8电流",
                         "大泵1电流", "大泵2电流", "小高速3电流", "小高速4电流", "小低速3电流", "小低速4电流"], axis=1,
                inplace=True)

        df["循泵方式"] = df["当前大泵数量"] * 100 + df["当前高速泵数量"] * 10 + df["当前低速泵数量"]
        df["循泵方式"] = df["循泵方式"].apply(lambda x: get_pump_type_by_int(int(x)))
        df["冷端耗功"] = df["当前大泵数量"] * 2200 + df["当前高速泵数量"] * 1381 + df["当前低速泵数量"] * 837 + df[
            "机力塔数量"] * 223
        df["净功率"] = df["全厂总功率"] * 1000 - df["冷端耗功"]  # kW
        df["气耗"] = df["天然气流量"] / df["净功率"]  # Nm3/(kWh)
        df.drop(columns=["当前大泵数量", "当前高速泵数量", "当前低速泵数量"], axis=1, inplace=True)

        df["主汽门前压力"] = df["主汽门前压力"] + df["大气压力"]
        df["高排压力"] = df["高排压力"] + df["大气压力"]
        df["中压主汽门前压力"] = df["中压主汽门前压力"] + df["大气压力"]
        df["低压蒸汽压力1"] = df["低压蒸汽压力1"] + df["大气压力"]
        df["低压蒸汽压力2"] = df["低压蒸汽压力2"] + df["大气压力"]
        df["热井出水压力"] = df["热井出水压力"] + df["大气压力"]

        df["主蒸汽焓"] = df.apply(lambda x: get_h_by_pt(x["主汽门前压力"], x["主汽门前温度"]), axis=1)
        df["高排焓"] = df.apply(lambda x: get_h_by_pt(x["高排压力"], x["高排温度"]), axis=1)
        df["主蒸汽流量"] = df["高压蒸汽流量1号炉"] + df["高压蒸汽流量2号炉"] + df["高压蒸汽减温水流量1号炉"] + df[
            "高压蒸汽减温水流量2号炉"]
        df["高压缸进汽能量"] = df["主蒸汽流量"] * df["主蒸汽焓"] * 1000
        合缸处漏气量 = 17.962
        df["高排流量"] = df["主蒸汽流量"] - 合缸处漏气量
        df["高排能量"] = df["高排焓"] * df["高排流量"] * 1000
        df["中压进汽流量"] = df["中压给水流量1号炉"] + df["中压给水流量2号炉"] + df["再热减温水流量1"] + \
                             df["再热减温水流量2"] + df["高排流量"] - df["FGH流量1号炉"] - df["FGH流量2号炉"]
        df["中压进汽焓"] = df.apply(lambda x: get_h_by_pt(x["中压主汽门前压力"], x["中压主汽门前温度"]),
                                    axis=1)
        df["中压进汽能量"] = df["中压进汽焓"] * df["中压进汽流量"] * 1000

        df.drop(columns=["主汽门前压力", "主汽门前温度", "高排压力", "高排温度", "主蒸汽焓", "高排焓", "主蒸汽流量",
                         "高排流量",
                         "中压给水流量1号炉", "中压给水流量2号炉", "再热减温水流量1", "再热减温水流量2", "FGH流量1号炉",
                         "FGH流量2号炉", "高压蒸汽减温水流量1号炉", "高压蒸汽减温水流量2号炉", "高压蒸汽流量1号炉",
                         "高压蒸汽流量2号炉", "中压主汽门前压力", "中压主汽门前温度", "燃机功率1", "燃机功率2"], axis=1,
                inplace=True)

        df["低压补汽能量1"] = df.apply(
            lambda x: get_h_by_pt(x["低压蒸汽压力1"], x["低压蒸汽温度1"]) * x["低压蒸汽流量1"] * 1000,
            axis=1)
        df["低压补汽能量2"] = df.apply(
            lambda x: get_h_by_pt(x["低压蒸汽压力2"], x["低压蒸汽温度2"]) * x["低压蒸汽流量2"] * 1000,
            axis=1)
        df["低压补汽能量"] = df["低压补汽能量1"] + df["低压补汽能量2"]

        df["热井出水流量"] = df["凝结水流量1号炉"] + df["凝结水流量2号炉"]
        df["热井出水焓"] = df.apply(lambda x: get_h_by_pt(x["热井出水压力"], x["热井出水温度"]), axis=1)
        df["热井出水能量"] = df["热井出水流量"] * df["热井出水焓"] * 1000

        df.drop(columns=["低压蒸汽压力1", "低压蒸汽温度1", "低压蒸汽流量1", "低压蒸汽压力2", "低压蒸汽温度2",
                         "低压蒸汽流量2",
                         "低压补汽能量1", "低压补汽能量2", "凝结水流量1号炉", "凝结水流量2号炉", "热井出水压力",
                         "热井出水温度", "热井出水流量", "热井出水焓", "中压进汽流量", "中压进汽焓"], axis=1,
                inplace=True)

        df["p"] = df["烟气热网加热器进水压力"] + df["大气压力"]
        df["热网加热器流量"] = df["烟气热网加热器进水流量"] / 2
        df["烟气换热量"] = df.apply(
            lambda x: (x["热网加热器流量"] * (get_h_by_pt(x["p"], x["烟气热网加热器供水温度1"]) - get_h_by_pt(x["p"], x[
                "烟气热网加热器进水温度1"]))
                       + x["热网加热器流量"] * (
                               get_h_by_pt(x["p"], x["烟气热网加热器供水温度2"]) - get_h_by_pt(x["p"], x[
                           "烟气热网加热器进水温度2"]))) * 1000,
            axis=1
        )

        df["供热能量"] = df["供热热量"] * 1000000 - df["烟气换热量"]
        df["轴功率"] = df["发电机有功功率_steam"] / 0.988 * 3600 * 1000
        df["凝汽器热负荷"] = df["高压缸进汽能量"] + df["中压进汽能量"] + df["低压补汽能量"] - df["高排能量"] - 18502790 \
                             - df["供热能量"] - df["轴功率"] - df["热井出水能量"]
        df["凝汽器热负荷"] = df["凝汽器热负荷"] / 3600 / 1000
        df.drop(columns=["供热热量", "供热能量", "发电机有功功率_steam", "轴功率", "高压缸进汽能量", "中压进汽能量",
                         "低压补汽能量", "高排能量", "供热能量", "热井出水能量", "大气压力", "热网加热器疏水流量1",
                         "热网加热器疏水流量2", "当前循泵运行方式", "当前风机运行方式", "凝汽器循环水进水温度1",
                         "凝汽器循环水进水温度2", "烟气热网加热器进水流量", "烟气热网加热器进水压力",
                         "烟气热网加热器进水温度1",
                         "烟气热网加热器供水温度1", "烟气热网加热器进水温度2", "烟气热网加热器供水温度2",
                         "热网加热器流量", "烟气换热量", "烟气热网加热器进水流量",
                         ], axis=1, inplace=True, errors='ignore')

        df["凝汽器热负荷"] = df["凝汽器热负荷"].rolling(window=10).mean()  # 因为凝汽器热负荷参数锯齿状波动严重，取移动平均值
        df["环境湿度"] = df["环境湿度"].rolling(10).mean()
        df = df[["全厂总功率", "凝汽器热负荷", "环境温度", "环境湿度", "循泵方式", "机力塔数量", "当前背压", "气耗"]]
        df = df.copy()
        df.drop(df.head(9).index, inplace=True)  # 删除前10行

        # 检测如果冷端运行方式改变，则删除后续的20分钟数据
        time_section = 20
        is_steady = ["F-top19"] * (time_section - 1)  # 前19行认为不稳定
        for row_idx in range(time_section, df.shape[0] + 1):
            _p = df.iloc[row_idx - time_section: row_idx][
                "循泵方式"]  # df.iloc[0, 20]["循泵方式"].all() -> df.iloc[19]=True
            _f = df.iloc[row_idx - time_section: row_idx]["机力塔数量"]
            if (_p.values == _p.values[0]).all() and (_f.values == _f.values[0]).all():
                is_steady.append("T")
            else:
                is_steady.append("F-冷端方式调整")

        # 考虑到需要保持时间的连续性，此处先标记该行数据需要删除，而不能直接对数据进行删除，否则会影响后续的稳定性条件判断

        # 检测功率如果波动超过20分钟内功率波动超过10MW，则删除数据
        for row_idx in range(time_section, df.shape[0] + 1):
            _p = df.iloc[row_idx - time_section:row_idx]
            if _p["凝汽器热负荷"].max() - _p["凝汽器热负荷"].min() > 10:
                is_steady[row_idx - 1] = "F-热负荷波动"  # 166
                # logger.debug(f"{row_idx=}, 热负荷max={_p['凝汽器热负荷'].max()},min={_p['凝汽器热负荷'].min()}")
            if _p["环境温度"].max() - _p["环境温度"].min() > 5:
                is_steady[row_idx - 1] = "F-环境温度波动"  # 39101
                # logger.debug(f"{row_idx=},环境温度max={_p['环境温度'].max()},min={_p['环境温度'].min()}")
            if _p["环境湿度"].max() - _p["环境湿度"].min() > 10:
                is_steady[row_idx - 1] = "F-环境湿度波动"  # 97
                # logger.debug(f"{row_idx=},环境湿度max={_p['环境湿度'].max()},min={_p['环境湿度'].min()}")

        df["steady"] = is_steady

        df = df[df["steady"].apply(lambda x: x.startswith("T"))]
        output_data.update({os.path.basename(file).split(".")[0]: df})

    write_excel("历史数据库.xlsx", output_data, index=True)


def load_history_file(file):
    """
    加载历史数据文件

    :param file:
    :return:
    """
    df = pd.DataFrame()
    if os.path.exists(file):
        if file.endswith(".csv"):
            df = pd.read_csv(file)
        else:
            df = read_excel(file, need_merge=True, index_col=0)
    return df


def find_condition(df, unit_condition, cold_condition: dict = None, auto_loose=False):
    """
    查找符合条件的历史工况，返回满足条件的历史数据及历史数据的统计平均值
    示例：
    df, df_mean = find_condition(file="历史数据库.xlsx",
                   unit_condition=[("凝汽器热负荷", 360, "1%"), ("环境温度", 27, "±2"), ("环境湿度", 50, "±10")],
                   cold_condition={"循泵方式": 'B', "机力塔数量": 8}
                   )

    :param auto_loose: 是否自动放宽条件限制，如果为True，当没有找到符合条件的工况时，会自动放宽允许的偏差为原来的两倍再次尝试，直到找到
    :param df: 历史数据dataframe，csv或者xlsx文件
    :param unit_condition: 机组的运行状态参数
    :param cold_condition: 冷端设备运行状态参数
    :return:
    """
    if df is None:
        return None
    origin_df = df.copy()
    # ------------------------------- 过滤冷端运行条件 -------------------------------------------------------
    cold_condition = {} if cold_condition is None else cold_condition  # 当不传入冷端运行条件时，直接跳过冷端条件限制
    for title, val in cold_condition.items():
        df = df[df[title] == val]
    if df.shape[0] == 0:
        logger.debug(f"历史上不存在{cold_condition}的冷端运行方式")
        return None
    # ------------------------------- 过滤冷端运行条件 -------------------------------------------------------

    for item in unit_condition:
        title, val, tolerance = item
        if is_number(val):
            if is_number(tolerance):
                val_min = val * (1 - tolerance)
                val_max = val * (1 + tolerance)
            elif tolerance.endswith("%"):
                tolerance = float(tolerance.replace("%", "")) / 100
                val_min = val * (1 - tolerance)
                val_max = val * (1 + tolerance)
            elif tolerance.startswith("±"):
                tolerance = float(tolerance.replace("±", ""))
                val_min = val - tolerance
                val_max = val + tolerance
            df = df[df[title] < val_max]
            df = df[df[title] > val_min]
    df = df.copy()
    if df.shape[0] < 5:
        if auto_loose:
            new_condition = []
            for item in unit_condition:
                title, val, tolerance = item
                if is_number(tolerance):
                    tolerance = tolerance * 2
                    if tolerance / val > 100:
                        return None
                elif tolerance.endswith("%"):
                    tolerance = float(tolerance.replace("%", "")) / 100
                    if tolerance > 10000:
                        return None
                    tolerance = f"{tolerance * 200}%"
                elif tolerance.startswith("±"):
                    tolerance = float(tolerance.replace("±", ""))
                    if tolerance / val > 100:
                        return None
                    tolerance = f"±{tolerance * 2}"
                new_condition.append((title, val, tolerance))
            # logger.debug(f"历史数据中满足条件的工况个数为{df.shape[0]}，放宽条件限制为：{new_condition}")
            return find_condition(origin_df, new_condition, cold_condition, auto_loose)
        else:
            return None
    else:
        # logger.debug(f"满足条件工况的平均参数为：{df.mean(numeric_only=True)}")
        return df


def generate_data(df):
    from yangke.base import interpolate_nd, kriging
    data = []
    points = []
    for Qc in range(100, 570, 20):
        for t0 in [-5, 0, 5, 10, 15, 20, 25, 30, 35]:
            for humid in [50]:
                for pump in ["A", "B", "C", "D", "E", "F", "G", "H", "I"]:
                    for fan in [3, 4, 5, 6, 7, 8]:
                        res = find_condition(df,
                                             unit_condition=[("凝汽器热负荷", Qc, "±10"), ("环境温度", t0, "±2.5"),
                                                             ("环境湿度", humid, "±100")],
                                             cold_condition={"循泵方式": pump, "机力塔数量": fan},
                                             auto_loose=False
                                             )
                        if res is not None:
                            mean = res.mean(numeric_only=True)
                            bp = mean["当前背压"]
                            gas = mean["气耗"]
                            data.append([mean["凝汽器热负荷"], mean["环境温度"], mean["环境湿度"], pump, fan, bp, gas])
                            print(f"{Qc=}, {t0=}, {humid=}, {pump=}, {fan=}, {bp=}, {gas=}")
                        else:
                            bp = None
                            gas = None

    res_df = pd.DataFrame(data=data,
                          columns=["凝汽器热负荷", "环境温度", "环境湿度", "循泵方式", "机力塔数量", "当前背压",
                                   "气耗"])

    for Qc in [100, 150, 200, 250, 300, 350, 400, 450, 500, 550]:
        for t0 in [-5, 0, 5, 10, 15, 20, 25, 30, 35]:
            for humid in [50]:
                for pump in ["A", "B", "C", "D", "E", "F", "G", "H", "I"]:
                    for fan in [3, 4, 5, 6, 7, 8]:
                        _df_slice = res_df[res_df["循泵方式"] == pump]
                        _df_slice = _df_slice[_df_slice["机力塔数量"] == fan]
                        if _df_slice.shape[0] < 5:
                            continue
                        _df_slice = _df_slice.copy()
                        points = _df_slice[["凝汽器热负荷", "环境温度", "当前背压"]]
                        input_data = points[["凝汽器热负荷", "环境温度"]].values.tolist()
                        output_data = points["当前背压"].values.tolist()
                        k = kriging(input_data, output_data)
                        val = k.predict([350, 25])
                        interpolate_nd(xi=[350, 25], dataframe=points, df_x_titles=["凝汽器热负荷", "环境温度"],
                                       df_y_title=["当前背压"])

    return res_df


def 删除非稳态数据(df: pd.DataFrame, title=None, slope=None, till=None, inplace=False):
    """
    删除非稳态数据，要求df中具有时间列，且时间必须连续，不能是间断的时间。
    分为两种情况。
    一种是某个参数波动幅度过大，需要剔除，则调用方法为：
    删除非稳态数据(df, "电功率", slope=5, till=20)  # 表示标记出20分钟内电功率波动超过5的数据条目

    另一种是某个参数突变后，删除后续的一段时间的数据，如循泵运行台数从2变为3时，删除后续20分钟数据，调用方法为：
    删除非稳态数据(df, "循泵运行台数", till=20)

    inplace参数表示是否真的删除具体数据，这在以下情况中是非常有用的，例如：
    当需要对同一组数据的"电功率", "循泵运行台数"等多个参数进行非稳态数据删除时，如果连续调用：
    df = 删除非稳态数据(df, "电功率", slope=5, till=20, inplace=True)
    df = 删除非稳态数据(df, "循泵运行台数", till=20, inplace=True)
    则第一次调用后，因为会删除一些数据，导致剩余的数据时间段并不连续，在进行后续判断非稳态删除操作时，导致非稳态判断错误，因此，在最后一句真实的
    删除操作前，所有的删除非稳态数据的方法调用都应该指定inpalce=False，只对需要删除的数据行做标记，而不是真正的删除，这样，后续的函数调用时，
    传入的df时间仍是连续的。
    合理的删除操作应为：
    df = 删除非稳态数据(df, "电功率1", slope=5, till=20, inplace=False) # 只标记，不删除
    df = 删除非稳态数据(df, "电功率2", slope=5, till=20, inplace=False) # 只标记，不删除
    df = 删除非稳态数据(df, "电功率3", slope=5, till=20, inplace=False) # 只标记，不删除
    df = 删除非稳态数据(df, "循泵运行台数", till=20, inplace=True) # 标记后，根据标记结果进行删除

    :param title:
    :param range:
    :param till:
    :return:
    """
    if "steady" in df.columns.tolist():
        has_steady = True
        col_idx = df.columns.tolist().index("steady")
    else:
        has_steady = False
        is_steady = ["F-top"] * (till - 1)  # 前19行认为不稳定，是一个19行的字符串列表
        col_idx = -1

    if slope is None:
        for row_idx in range(till, df.shape[0] + 1):
            运行方式 = df.iloc[row_idx - till: row_idx][title]
            if (运行方式.values == 运行方式.values[0]).all():  # 如果持续时间内，运行方式都不变
                if not has_steady:
                    is_steady.append("T")
                else:
                    pass  # 如果有steady列，则不进行操作
            else:
                if not has_steady:
                    is_steady.append("F-冷端方式调整")
                else:
                    df.iat[row_idx, col_idx] = "F-冷端方式调整"
    else:
        # 检测功率如果波动超过20分钟内功率波动超过10MW，则删除数据
        for row_idx in range(till, df.shape[0] + 1):
            _p = df.iloc[row_idx - till:row_idx]
            if _p[title].max() - _p[title].min() > slope:
                if not has_steady:
                    is_steady.append(f"F-{title}波动")
                else:
                    df.iat[row_idx - 1, col_idx] = f"F-{title}波动"
            else:
                if not has_steady:
                    is_steady.append("T")
                else:
                    pass
    if not has_steady:
        df["steady"] = is_steady
    if inplace:
        df = df[df["steady"].apply(lambda x: x.startswith("T"))]
    return df.copy()


def beijingshenhua():
    """
    生成北京神华国华机组的背压计算数据库
    :return:
    """
    des1 = TagsRead.get_keys()
    tags1 = TagsRead.get_values()

    des2 = TagsWrite.get_keys()
    tags2 = TagsWrite.get_values()

    tags1.extend(tags2)
    des1.extend(des2)
    df = load_history_file("历史数据库.xlsx")
    res = generate_data(df)
    res.to_csv("结果3.csv", index=False)


if __name__ == "__main__":
    tags = TagsRead.get_values()
    des = TagsRead.get_keys()
    yyyy, mm, dd, hour, minute, second = (2022, 1, 1, 0, 0, 0)
    start_time = datetime.datetime(year=yyyy, month=mm, day=dd, hour=hour, minute=minute, second=second)
    yyyy, mm, dd, hour, minute, second = (2022, 1, 1, 23, 59, 59)
    end_time = datetime.datetime(year=yyyy, month=mm, day=dd, hour=hour, minute=minute, second=second)
    df = get_history(tags, des, start_time=start_time, end_time=end_time)



import json
import requests
from yangke.web.baidu import get_token


def ask(question: str):
    """
    一个思知的对话机器人，但貌似准确度很低
    :param question:
    :return:
    """
    sess = requests.get('https://api.ownthink.com/bot?spoken={}'.format(question))
    answer = sess.text
    answer = json.loads(answer)
    return answer


def ask_baidu(question: str):
    """
    百度的对话机器人，详情参见百度 机器人对话API文档，https://ai.baidu.com/ai-doc/UNIT/qk38gggxg

    Unit主页 https://ai.baidu.com/unit/home

    :param question:
    :return: 问题答案的列表，每一个列表都对应一个可能的答案
    """

    access_token = get_token(client_id="fmt3q4GOHR5lyzjGb78I9Otr", client_secret="DHcIclrolnmPnMFOyf6mjnZ6V2Qda76o")
    url = 'https://aip.baidubce.com/rpc/2.0/unit/service/chat?access_token=' + access_token  # 机器人对话API，沙盒环境
    # url = 'https://aip.baidubce.com/rpc/2.0/unit/bot/chat?access_token=' + access_token  # 技能对话API

    post_data = {
        "log_id": "7758521",
        "version": "2.0",
        # "service_id": "S10000",
        "skill_ids": ["1033038"],
        "session_id": "",
        "request": {
            "query": question,
            "user_id": "UNIT_DEV_YK"
        },
        "dialog_state": {"contexts": {"SYS_REMEMBERED_SKILLS": ["1033038"]}}
    }
    post_data = str(post_data).replace("'", '"').encode('utf-8')

    headers = {'content-type': 'application/x-www-form-urlencoded'}
    response = requests.post(url, data=post_data, headers=headers)
    result = []  # 答案的列表
    if response:
        print(response.json())
        response_list = response.json().get('result').get('response_list')
        for res in response_list:
            action = res.get('action_list')[0]  # action_list是一个列表，这里去列表第一项，如果以后遇到多答案的问题，再更新
            result.append(action.get('say'))

    return result


# def glove():

if __name__ == "__main__":
    import re

    qingming_dict = {}
    for year in range(1990, 2100):
        ans = ask_baidu(f'{year}年清明节是几月几号')[0]
        ans = re.findall(".*月(.)日.*", ans)
        while len(ans) == 0:
            ans = ask_baidu(f'{year}年清明节是几月几号')[0]
            ans = re.findall(".*月(.)日.*", ans)
        qingming_dict[str(year)] = ans[0]
    print(qingming_dict)



import copy
import random
import re
from collections import OrderedDict
from typing import Optional

import numpy as np
import optuna
import pandas as pd
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, Dataset
from torch.utils.tensorboard import SummaryWriter

from yangke.base import get_settings, YkDict, is_number
from yangke.common.config import logger
from yangke.common.fileOperate import read_csv_ex
import os

os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"


class PriorLoss(nn.Module):
    """
    自定义的损失函数，可以使用已有的先验知识
    """

    def __init__(self, input_titles, net):
        self.input_titles = input_titles
        super(PriorLoss, self).__init__()
        self.rules = []
        self.reg_rules = []  # [("der>0, '电功率'")]  # 表示输出对电功率的导数大于0
        self.net = net  # 损失函数所属的神经网络
        self.punish = 1000  # 当预测结果违背先验知识时的损失函数惩罚因子

    def forward(self, pred, target, input):
        """
        损失函数返回的参数是0-dim的tensor

        :param pred: 输出参数预测值
        :param target: 输出参数实际值
        :param input: 输入参数
        :return:  2866343884064  2866343893200
        """
        x = input
        x.requires_grad = True
        if x.grad is not None:
            x.grad = x.grad - x.grad  # 清空以前的梯度
        predict = self.net(x)
        # y对x求导
        _y = predict.mean()
        _y.backward()  # 只是计算梯度，不进行权值更新
        grad = x.grad  # 则此处x.grad就是y对x的导数
        x.requires_grad = False

        correction = self.apply_rules(x, grad)  # 先验知识的修正项

        ret = torch.abs(pred - target).squeeze()
        ret = torch.pow(ret, 2) + correction
        loss = ret.mean()
        return loss

    def apply_rules(self, input, grad):
        """
        在【输入参数】和【输出对输入参数的梯度】上求解先验知识的修正项
        :param input: Tensor:(32,2)当batch_size=32，input_dims=2
        :param grad: Tensor:(32,2)
        :return:
        """
        correction = torch.zeros_like(grad[:, 0])  # 无论输入有多少个，损失函数值只有一个，因此只取grad的第一个为维度
        for rr in self.reg_rules:
            if rr[0] == "der>0":
                idx = self.input_titles.index(rr[1])
                _cor = grad[:, idx]  # 所有输入对x的导数，该数应该大于0才符合先验知识
                _cor = -_cor * self.punish  # 如果导数小于0，则correction为大于0的数，即损失增大
                _cor = torch.clamp(_cor, 0, None)  # 将所有小于0的数用0代替，即导数大于0时，损失修正值为0
                correction = correction + _cor
            elif rr[0] == "der<0":
                idx = self.input_titles.index(rr[1])
                _cor = grad[:, idx]
                _cor = _cor * self.punish  # 如果导数大于0，则correction为大于0的数，即损失增大
                _cor = torch.clamp(_cor, 0, None)
                correction = correction + _cor
            else:
                logger.error(f"暂不支持其他条件！{rr}")
                return
        return correction

    def translate_rules(self):
        for rule in self.rules:
            if rule.get("isChecked"):
                expression = rule.get("expression")

            else:
                op = rule.get("op")
                if op == "任何条件":
                    expression = rule.get("expression")
                else:
                    logger.error(f"暂不支持带条件的先验知识")
                    expression = None
            res = re.findall('\[(.+?)\]', expression)  # 加一个?则是最小匹配，否则时贪婪匹配
            plus = re.findall('.*单调递(.+)', expression)
            if len(res) == 2 and len(plus) == 1:
                if plus[0] == "增":
                    self.reg_rules.append(['der>0', res[1]])
                else:
                    self.reg_rules.append(['der<0', res[1]])
            else:
                logger.error(f"未知的条件格式")

    def add_knowledge(self, rule):
        """
        添加已有的先验知识规则。示例：
        1. 任何条件下，背压都随电功率单调递增
        {'isChecked': False, 'op': '任何条件', 'expression': '[背压]随[电功率]单调递增'}

        2. 电功率小于100时，背压随电功率单调递增
        {'isChecked': False, 'var1': '电功率', 'op': '小于', 'var2': '100', 'expression': '[背压]随[电功率]单调递增'}

        3. 同时传入多个先验知识的规则，使用列表表示，如下：
        [{'isChecked': False, 'var1': '电功率', 'op': '任何条件', 'var2': '电功率', 'expression': '[背压]随[电功率]单调递增'}, {'isChecked': False, 'var1': '电功率', 'op': '任何条件', 'var2': '电功率', 'expression': '[背压]随[环境温度]单调递增'}]

        4. 使用表达式表示规则，则isChecked设置为True
        {'isChecked': True, "expression": "[背压]随[电功率]单调递增 if [电功率]>100 or [电功率]<500"}

        y随着输入1的增加而增加，则
        rule = 'x(1).grad>0' 或 rule = 'x(电功率).grad>0'  其中第一个输入参数名为电功率

        :param rule:
        :return:
        """
        if isinstance(rule, list):
            self.rules = rule
        else:
            self.rules.append(rule)

        self.translate_rules()

    def clear_knowledge(self):
        self.rules = []


def standard_reverse(out, mean, std):
    """
    将预测结果反标准化

    :param out:
    :return:
    """
    if isinstance(std, pd.Series):
        std = std.item()  # Series.values是ndarray, item()返回的是数值float型
    if isinstance(mean, pd.Series):
        mean = mean.item()

    if isinstance(out, torch.Tensor):
        out = out.detach().numpy()
    return out * std + mean


def all_number(row: pd.Series):
    """
    判断row中的值是否都是数值

    :param row:
    :return:
    """
    res = True
    for v in list(row.to_numpy()):
        if not is_number(v):
            res = False
            break
    return res


class DataSetFitting(Dataset):
    def __init__(self, data_source, x_title=None, y_title=None, mean=None, std=None, proportion_train=0.8,
                 proportion_test=0.2, classify=False):
        """
        构建数据集。
        用法示例：
        # 1. 构建数据集
        dataset = DataSetFitting(df, ["x1", "x2"], ["y"])
        # 2. 数据集归一化
        dataset.set_standard(normal_method="min-max")

        :param data_source: 数据源，可以为settings.yaml文件配置，也可以是pd.DataFrame对象
        :param x_title: 输入参数的列标题，如果从配置文件加载，则可以为None
        :param y_title: 输出参数的列标题，如果从配置文件加载，则可以为None
        :param mean: 数据集的平均值
        :param std: 数据集的标准差
        :param proportion_train: 训练集的比例，默认80%
        :param proportion_test: 测试集比例，默认20%
        :param classify: 是否为分类问题的数据集，如果是分类问题，则归一化不会对标签应用，否则会将标签也归一化处理
        """
        self.dataframe = None
        self.x_title = None
        self.y_title = None
        self.normal_method = None
        self.y_range = None
        self.x_range = None
        self.proportion_train = proportion_train
        self.proportion_test = proportion_test
        if isinstance(data_source, pd.DataFrame):
            self.dataframe = data_source
            self.x_title = x_title or []
            self.y_title = y_title or []
            if not isinstance(self.y_title, list):
                self.y_title = [self.y_title]
        elif isinstance(data_source, YkDict):
            files = data_source.get_settings("dataset.data_file")
            self.normal_method = data_source.get_settings("dataset.normalization") or "z-score"
            input_set = data_source.get("input")
            self.x_title = [item["name"] for item in input_set]
            self.x_range = [None if item.get("range") is None else eval(item["range"]) for item in input_set]
            del input_set
            output_set = data_source.get_settings("output.para")
            if len(output_set) > 0 and isinstance(output_set[0], str):
                self.y_title = output_set
                self.y_range = None
            else:
                self.y_title = [item["name"] for item in output_set]
                self.y_range = [None if item.get("range") is None else eval(item["range"]) for item in output_set]
            for file in files:
                ext = file.get("type")
                filename = file.get("name")
                if not os.path.exists(filename):
                    continue
                if ext == "csv":
                    data_frame = read_csv_ex(filename)
                else:
                    data_frame = pd.read_excel(filename)
                if self.dataframe is not None and data_frame is not None:
                    self.dataframe = pd.concat([self.dataframe, data_frame], axis=0, ignore_index=True)
                else:
                    self.dataframe = self.dataframe or data_frame

            # 删除某些行
            self.drop_by_condition(data_source.get_settings("dataset.drop"))

        self.classify = classify
        # 仅保留x,y相关的数据列，删除其他列
        titles = self.x_title.copy()
        titles.extend(self.y_title)
        if self.dataframe.shape[0] < 1:
            logger.warning("传入的数据文件中数据为空或数据文件不存在")
        self.dataframe = self.dataframe[titles].copy()

        # 找出dataframe中非数值的行
        self.dataframe["is_number"] = self.dataframe.apply(lambda _: all_number(_), axis=1)
        origin_titles = set(self.dataframe.index)
        self.dataframe = self.dataframe[self.dataframe['is_number']]
        self.non_numeric_lines = list(origin_titles - set(self.dataframe.index))
        if len(self.non_numeric_lines) > 0:
            logger.warning(f"数据集中存在非数值的行，已删除这些行，行号为：{self.non_numeric_lines}")
        self.dataframe = self.dataframe.astype(float).copy()
        self.dataframe = self.dataframe[titles].copy()

        # 数据标准化
        self.mean = mean  # 可能为None，后续处理
        self.std_range = std
        self.dataframe_std = None  # 归一化的数据集
        # self.set_standard()

    def set_standard(self, mean=None, std_range=None, normal_method=None, x_range=None, y_range=None):
        """
        设置数据的归一化参数，在初始化部分数据时，直接从部分数据集得到的归一化参数可能出现偏差，这里提供一种外部设置的方法。
        不建议中途更改数据归一化方法，否则可能导致模型预测结果很离谱。

        min-max将数据归一化到(-1,1)的区间上

        :param mean: 参数的平均值
        :param std_range: 参数的标准差或振幅
        :param normal_method: 数据归一化方法，支持"z-score"和"min-max"
        :param y_range: [(0, 20)]
        :param x_range: [(0, 100), (0, 20), (0, 30)]
        :return:
        """
        if normal_method is None or normal_method == self.normal_method:
            if mean is None and std_range is None:
                logger.debug(f"mean and std is None, try to calculate by method {self.normal_method}")
                if self.normal_method == "z-score":
                    self.mean = self.mean or self.dataframe.mean()
                    self.std_range = self.std_range or self.dataframe.std()
                    self.dataframe_std = self.standard(self.dataframe)  # 必须放在这里，不然可能由于编译优化被忽略
                    if self.classify:  # 分类问题，不能对标签进行归一化处理
                        self.dataframe_std[self.y_title] = self.dataframe[self.y_title]
                    return
                else:
                    self.x_range = x_range or self.x_range
                    self.y_range = y_range or self.y_range
                    if self.y_range is not None:
                        x_min = [x[0] for x in self.x_range]
                        y_min = [y[0] for y in self.y_range]
                        x_max = [x[1] for x in self.x_range]
                        y_max = [y[1] for y in self.y_range]
                        x_min.extend(y_min)
                        all_min = np.asarray(x_min)
                        x_max.extend(y_max)
                        all_max = np.asarray(x_max)
                        title = copy.deepcopy(self.x_title)
                        title.extend(self.y_title)
                        self.mean = pd.Series(data=(all_min + all_max) / 2, index=title)
                        self.std_range = pd.Series(data=(all_max - all_min) / 2, index=title)
                    else:
                        logger.warning("数据归一化方法为min-max，但未设置各参数的取值范围")
            else:
                self.mean = mean
                self.std_range = std_range
            if self.mean is not None and self.std_range is not None:
                self.dataframe_std = self.standard(self.dataframe)
            else:
                logger.debug(f"数据归一化失败，原因：{self.mean=}, {self.std_range=}")
                self.dataframe_std = None
        else:
            self.normal_method = normal_method
            self.set_standard(mean=mean, std_range=std_range, normal_method=normal_method, x_range=x_range,
                              y_range=y_range)
        if self.classify:  # 分类问题，不能对标签进行归一化处理
            self.dataframe_std[self.y_title] = self.dataframe[self.y_title]

    def standard(self, df):
        """
        数据标准化

        :param df: Dataframe(num, input_and_output_dim)
        :return:
        """
        # Dataframe和Series运算的前提是：Dataframe的列标题和Series的索引名必须一一对应，而不仅仅看维度，否则可能会出错，
        return (df - self.mean) / self.std_range  # mean Series (input_and_output_dim,)

    def standard_reverse(self, out):
        """
        将预测结果反标准化

        :param out:
        :return:
        """
        std = self.std_range[self.y_title]
        mean = self.mean[self.y_title]
        if isinstance(std, pd.Series):
            std = std.item()  # Series.values是ndarray
        if isinstance(mean, pd.Series):
            mean = mean.item()
        return out * std + mean

    def __getitem__(self, index):
        """
        DataSet子类必须实现的方法，用于根据索引返回一条数据，数据类型需要是Tensor

        :param index:
        :return:
        """
        if self.dataframe_std is None:
            logger.warning(f"数据集没有归一化处理")
            self.dataframe_std = self.dataframe
        single_item = self.dataframe_std.iloc[index, :]
        _x = torch.from_numpy(single_item[self.x_title].to_numpy()).to(torch.float32)
        if self.classify:  # 如果是分类问题，则输出必须是torch.LongTensor类型
            # crossEntropyLoss会将整数转换成one-hot类型的数组，因此每一个标签必须是整数，不能是数组
            _y = torch.from_numpy(single_item[self.y_title].to_numpy()).to(torch.long).squeeze()
            # _y = _y.long()
        else:
            _y = torch.from_numpy(single_item[self.y_title].to_numpy()).to(torch.float32)

        return _x, _y

    def __len__(self):
        """
        DataSet子类必须实现的方法，用于获取DataSet的大小

        :return:
        """
        return self.dataframe.shape[0]

    def get_size(self):
        return self.dataframe.shape[0]

    def drop_by_condition(self, conditions):
        for cond in conditions:
            if list(cond.keys())[0] == "or":
                for co in cond.get("or"):
                    if "<=" in co:
                        title, value = tuple(co.split("<="))
                        title = str(title).strip()
                        value = float(value)
                        self.dataframe = self.dataframe[self.dataframe[title] > value]  # 删除小于的行 = 保留大于的行
                    elif ">=" in co:
                        title, value = tuple(co.split(">="))
                        title = str(title).strip()
                        value = float(value)
                        self.dataframe = self.dataframe[self.dataframe[title] < value]  # 删除小于的行 = 保留大于的行
                    elif "<" in co:
                        title, value = tuple(co.split("<"))
                        title = str(title).strip()
                        value = float(value)
                        self.dataframe = self.dataframe[self.dataframe[title] >= value]  # 删除小于的行 = 保留大于的行
                    elif ">" in co:
                        title, value = tuple(co.split(">"))
                        title = str(title).strip()
                        value = float(value)
                        self.dataframe = self.dataframe[self.dataframe[title] <= value]  # 删除大于的行 = 保留小于的行

    def split_set(self, part1=None, part2=None, part3=None):
        """
        按照指定的比例将数据集分割，一般用于将总体数据集分割为训练集，测试集，验证集等

        :param part1:
        :param part2:
        :param part3:
        :return:
        """
        part1 = part1 or self.proportion_train
        part2 = part2 or self.proportion_test
        if part3 is not None:
            assert part1 + part2 + part3 == 1, "数据集比例之和不为1"
            size = self.get_size()
            size1, size2 = int(part1 * size), int(part2 * size)
            set1, set2, set3 = torch.utils.data.random_split(self, [size1, size2, size - size1 - size2])
            return set1, set2, set3
        elif part2 is not None:
            if part1 + part2 < 1:
                return self.split_set(part1, part2, 1 - part1 - part2)
            else:
                size = self.get_size()
                size1 = int(part1 * size)
                return torch.utils.data.random_split(self, [size1, size - size1])
        else:
            size = self.get_size()
            size1 = int(part1 * size)
            return torch.utils.data.random_split(self, [size1, size - size1])


class DataFitterNet(torch.nn.Module):
    def __init__(self, settings1, trial=None):
        """
        一个用于数据拟合的神经网络类库，神经网络架构通过settings.yaml文件进行配置

        :param settings1:
        :param trial: 使用optuna超参数调优时会传入该参数
        """
        super(DataFitterNet, self).__init__()
        self.dataset: DataSetFitting | None = None  # 数据集的信息，主要需要使用数据集中的归一化与反归一化的信息，但不会保存dataframe数据
        cfg = settings1.get("networks")  # 获取神经网络结构信息
        settings1 = YkDict(settings1)
        self.settings = settings1
        self.trial = trial
        self.cfg = cfg
        self.in_num = 0  # 输入层神经元个数，对应输入参数的个数
        self.out_num = 1
        loss_func_type = settings1.get_settings("networks.loss_fn.type")
        if loss_func_type == "CrossEntropy":  # 必须是分类问题才能使用CrossEntropy损失函数
            self.classify = True
            self.class_num = settings1.get_settings("networks.loss_fn.bak_info")

        train_settings = settings1.get_settings("networks.train") or {}
        self.lr = self.get_para_of_optuna_str(train_settings.get("learning_rate") or 1e-3, "lr")
        self.epochs = int(train_settings.get("epochs") or 10)
        self.batch_size = self.get_para_of_optuna_str(train_settings.get("batch_size") or 64, "batch_size")

        self.mean: Optional[pd.Series] = None
        self.std: Optional[pd.Series] = None
        self.normal_method = None  # 数据归一化方法
        self.x_title = settings1.get("input") or []
        self.y_title = settings1.get("output") or []
        if not isinstance(self.y_title, list):  # 因为输出经常为一个参数，如果用户没有以列表形式指定该参数，则转换成列表形式
            self.y_title = [self.y_title]
        self.max_err = 0
        self.average_err = 0
        _cell_num_last_layer = 0  # 循环中记录上一层神经网络的输出个数
        i = 1
        layer_dict = OrderedDict()
        for layer in cfg.get("layers"):  # 从配置中拿到神经网络各层的信息，进而构建神经网络
            _type = self.get_para_of_optuna_str(layer.get("type"), f"layer_type{i}")
            _cell_num = 1
            if _type == "input":  # 输入/输出层的神经元个数可以根据输入/输出参数个数自动确定
                _cell_num = layer.get("cell_num") or 10
                if layer.get("cell_num") == "auto" or layer.get("cell_num") is None:
                    self.in_num = len(settings1.get("input"))
                    _cell_num = self.in_num
                else:
                    self.in_num = int(layer.get("cell_num") or 10)
                    _cell_num = self.in_num
            elif _type == "linear":
                bias = False if layer.get("bias") is False else True
                _cell_num = self.get_para_of_optuna_str(layer.get("cell_num") or 10, f"cell_num{i}")
                layer_dict[f"layer{i}"] = nn.Linear(_cell_num_last_layer, _cell_num, bias=bias)
            elif _type == "relu":
                layer_dict[f"layer{i}"] = nn.ReLU()
                _cell_num = _cell_num_last_layer  # 激活函数输入输出个数不变
            # elif _type == 'softmax':
            #     self.layers.append(nn.Softmax())
            elif _type == "sigmoid":
                layer_dict[f"layer{i}"] = nn.Sigmoid()
                _cell_num = _cell_num_last_layer
            elif _type == "tanh":
                layer_dict[f"layer{i}"] = nn.Tanh()
                _cell_num = _cell_num_last_layer
            elif _type == "dropout":
                p = self.get_para_of_optuna_str(layer.get('rate'), "dropout_rate")
                layer_dict[f"layer{i}"] = nn.Dropout(p=p)
                _cell_num = _cell_num_last_layer
            elif _type.lower() == "rbf":
                _cell_num = layer.get("cell_num") or 10
                self.in_num = len(settings1.get("input"))
                self.out_num = len(settings1.get("output"))
                logger.error("暂不支持径向基神经网络")
            elif _type.lower() == "none":  # 空层
                continue
            elif _type == "output":  # 输入/输出层的神经元个数可以根据输入/输出参数个数自动确定
                if self.classify:  # 如果是分类问题，则输出个数为类别数量，分别对应每一类的概率
                    self.out_num = self.class_num
                else:
                    if layer.get("cell_num") == "auto" or layer.get("cell_num") is None:
                        self.out_num = len(self.y_title)
                    elif settings1.get_settings("output.type") == "single_output":
                        if layer.get("cell_num") != "auto" and int(layer.get("cell_num")) != 1:
                            logger.warning(
                                "数据集使用single_output，但神经网络模型输出层单元个数不为1，忽略神经网络输出单元个数设置！")
                        self.out_num = 1
                    else:
                        self.out_num = layer.get("cell_num")
                bias = False if layer.get("bias") is False else True
                layer_dict[f"layer{i}"] = nn.Linear(_cell_num_last_layer, self.out_num, bias=bias)
            else:
                logger.warning(f"发现不支持的layer类型：{_type}")

            _cell_num_last_layer = _cell_num
            i = i + 1

        if self.classify:
            layer_dict[f"layer{i}"] = nn.Softmax(dim=1)  # nn.functional.softmax()

        self.net = nn.Sequential(layer_dict)
        weight_decay = self.get_para_of_optuna_str(settings1.get_settings("networks.optimizer.weight_decay") or 0,
                                                   "weight_decay")
        if settings1.get_settings("networks.optimizer.type") == "adam":
            self.optimizer = torch.optim.Adam(self.parameters(), lr=self.lr, weight_decay=weight_decay)
        else:
            self.optimizer = torch.optim.SGD(self.parameters(), lr=self.lr, weight_decay=weight_decay)

        if settings1.get_settings("networks.loss_fn.type") == "MSE":
            self.loss_fn = nn.MSELoss(reduction="mean")
        elif settings1.get_settings("networks.loss_fn.type").lower() == "max":
            from yangke.stock.prediction.pytorch_pred_stock import MAXLoss
            self.loss_fn = MAXLoss()
        elif settings1.get_settings("networks.loss_fn.type") == "PriorLoss":
            self.loss_fn = PriorLoss(self.x_title, self.net)
            self.loss_fn.add_knowledge(settings1.get_settings("networks.loss_fn.bak_info"))
        elif settings1.get_settings("networks.loss_fn.type") == "CrossEntropy":
            self.loss_fn = nn.CrossEntropyLoss()
        elif settings1.get_settings("networks.loss_fn.type") == "BCE":
            self.loss_fn = nn.BCELoss()
        elif settings1.get_settings("networks.loss_fn.type") == "CTC":
            self.loss_fn = nn.CTCLoss()
        else:
            self.loss_fn = nn.GaussianNLLLoss

    def forward(self, x):
        out = self.net(x)
        return out

    def prediction(self, x_input, is_normalized=False):
        """
        预测指定输入参数，输入参数应为未进行归一化处理的原始数据

        :param x_input: x.shape = torch.Size([batch_size, input_dim])，也可以是长度为input_dim的列表或ndarray
        :param is_normalized: x_input是否已经归一化，默认为False，则prediction会执行归一化，否则不执行归一化直接预测
        :return:
        """
        mean = self.mean
        std = self.std
        if isinstance(self.mean, pd.Series):
            mean = self.mean[self.x_title].values
        if isinstance(self.std, pd.Series):
            std = self.std[self.x_title].values

        if isinstance(x_input, pd.DataFrame):
            x_input = x_input[self.x_title].values  # df.values是一个ndarray对象
        elif isinstance(x_input, list):
            x_input = np.asarray(x_input)

        # 将x_input转换成torch.tensor类型
        if isinstance(x_input, np.ndarray):
            if not is_normalized:
                x_input = (x_input - mean) / std  # 输入数据归一化
            x_input = torch.from_numpy(x_input)
        x_input = x_input.to(torch.float32)
        self.eval()
        y = self.forward(x_input)
        if self.classify:  # 分类问题，需要找到输出中概率最大的数的索引为最终的类别
            res = torch.argmax(y, dim=1)
        else:
            res = standard_reverse(y, mean=self.mean[self.y_title], std=self.std[self.y_title])
            if isinstance(res, np.ndarray) and res.size == 1:
                res = res.item()
        return res

    def set_standard(self, mean, std, x_titles, y_titles, normal_method):
        """
        设置预测时用到的归一化数据的信息

        :param mean:
        :param std:
        :param x_titles:
        :param y_titles:
        :param normal_method:
        :return:
        """
        self.mean = mean
        self.std = std
        self.x_title = x_titles
        self.y_title = y_titles
        self.normal_method = normal_method

    def save_yk(self, path):
        checkpoint = {
            'normal_method': self.normal_method,
            'mean': self.mean,
            'std': self.std,
            'x_title': self.x_title,
            'y_title': self.y_title,
            'model_state_dict': self.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
            'max_err': self.max_err,
            'average_err': self.average_err,
            'settings': self.settings
        }
        torch.save(checkpoint, path)

    @staticmethod
    def load_yk(path):
        if path is None or not os.path.exists(path):
            logger.debug("神经网络模型文件不存在！")
            return None
        checkpoint = torch.load(path)
        model1 = DataFitterNet(checkpoint.get("settings"))
        model1.set_standard(checkpoint['mean'], checkpoint['std'], checkpoint['x_title'], checkpoint['y_title'],
                            checkpoint['normal_method'])
        model1.max_err = checkpoint['max_err']
        model1.average_err = checkpoint['average_err']
        model1.load_state_dict(checkpoint['model_state_dict'])
        model1.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        return model1

    def __str__(self):
        str1 = super(DataFitterNet, self).__str__()
        return str1[:-1] + "  (Optimizer): " + self.optimizer.__str__().replace("\n", "\n  ", 1).replace("\n",
                                                                                                         "\n  ") + "\n)"

    def device(self):
        return next(self.parameters()).device

    def train_loop(self, data_loader):
        size = len(data_loader.dataset)
        self.train()  # 设置神经网络为训练模式
        loss = 0
        is_tqdm = self.settings.get_settings("print.tqdm")
        if is_tqdm:
            from tqdm import tqdm
            pb = tqdm(data_loader)
        else:
            pb = data_loader
            pb.set_description = logger.info
        relative_err = 0  # 如果是分类问题，则记录的是分类错误的数据条目数量
        max_err = 0
        for batch, (x, y) in enumerate(pb):  # 这里的(x, y)是归一化后的数据
            x = x.to(self.device())
            y = y.to(self.device())
            predict = self.forward(x)

            if isinstance(self.loss_fn, PriorLoss):  # 先验知识的损失函数需要y对x的导数
                loss = self.loss_fn(predict, y, x)
            elif isinstance(self.loss_fn, nn.MSELoss):
                loss = self.loss_fn(predict, y)
            elif isinstance(self.loss_fn, nn.CrossEntropyLoss):
                loss = self.loss_fn(predict, y)
            elif isinstance(self.loss_fn, nn.CTCLoss):
                logger.error(f"暂不支持CTC损失函数")
                return
            else:
                logger.error(f"暂不支持{self.loss_fn.__class__.__name__}损失函数")

            self.optimizer.zero_grad()  # 每一次训练梯度要清零
            loss.backward()
            self.optimizer.step()

            if batch % 10 == 0:
                # loss, current = loss.parent(), batch * len(x)  # loss是Tensor对象，没有parent，使用tensor.item()取值
                loss, current = loss.item(), batch * len(x)  # loss是Tensor对象，没有parent，使用tensor.item()取值
                pb.set_description(f"loss: {loss:>7f}  [{current:>6d}/{size:>6d}-{self.y_title}]")

            if self.classify:
                relative_err += torch.sum(y.flatten() != torch.argmax(predict, dim=1))
            else:
                # 累加当前的平均误差，更新当前的最大误差
                pred = self.dataset.standard_reverse(predict)  # 预测的真实值，任何情况都可以计算
                _y = self.dataset.standard_reverse(y)  # 实际的真实值
                max_err = max(max_err, (abs(pred - _y) / abs(_y)).max())  #
                relative_err = ((pred - _y).abs() / _y.abs()).sum() + relative_err  # 所有批次总的相对误差
        if self.classify:
            relative_err = relative_err / size
            logger.debug(f"训练集测试结果：分类错误概率为{relative_err}")
        else:
            relative_err = relative_err / size  # 平均相对误差
            logger.debug(f"训练集预测结果：相对误差为{relative_err.item():>8f}，最大误差为{max_err.item():>8f}")
        return loss, (relative_err, max_err)

    def test_loop(self, data_loader):
        num_batches = len(data_loader)
        size = len(data_loader.dataset)
        test_loss = 0
        self.eval()  # 停止更新权值参数，dropout层全部激活
        relative_err = 0  # 分类问题，则为分类错误的数据条目的个数
        max_err = 0
        with torch.no_grad():  # 停止autograd模块的工作，不影响dropout和batch norm层的行为，可以和self.eval()同时使用
            for _x, _y in data_loader:
                _x = _x.to(self.device())
                _y = _y.to(self.device())
                pred = self.forward(_x)

                # 不同的损失函数，需要传入的参数不一样
                if isinstance(self.loss_fn, PriorLoss):  # 先验知识的损失函数需要y对x的导数
                    # PriorLoss不支持torch.no_grad下计算损失
                    # loss = self.loss_fn(pred, _y, _x)
                    ...
                elif isinstance(self.loss_fn, nn.MSELoss):
                    loss = self.loss_fn(pred, _y)
                    test_loss += loss.item()  # 所有批次总的损失函数值
                elif isinstance(self.loss_fn, nn.CrossEntropyLoss):
                    loss = self.loss_fn(pred, _y)
                    test_loss += loss.item()  # 所有批次总的损失函数值
                elif isinstance(self.loss_fn, nn.CTCLoss):
                    logger.error(f"暂不支持CTC损失函数")
                    return
                else:
                    logger.error(f"暂不支持{self.loss_fn.__class__.__name__}损失函数")

                if self.classify:
                    relative_err += torch.sum(_y.flatten() != torch.argmax(pred, dim=1))
                else:
                    pred = self.dataset.standard_reverse(pred)  # 预测的真实值
                    _y = self.dataset.standard_reverse(_y)  # 实际的真实值
                    # max_err = max(max_err, (abs(pred-y)/abs(y)).max(0))  # tensor.max()返回最大值,tensor.max(0)返回最大值和索引
                    max_err = max(max_err, (abs(pred - _y) / abs(_y)).max())  #
                    relative_err = ((pred - _y).abs() / _y.abs()).sum() + relative_err  # 所有批次总的相对误差

        relative_err = relative_err / size  # 平均相对误差
        if isinstance(self.loss_fn, PriorLoss):
            logger.debug(f"测试集预测结果({self.y_title[0]})")
        elif self.classify:
            logger.debug(f"测试集预测结果：分类错误的概率为：{relative_err}")
        else:
            test_loss /= num_batches  # 平均损失
            logger.debug(f"测试集预测结果({self.y_title[0]}): 平均损失为{test_loss:>8f}")
        return loss, (relative_err, max_err)

    def validate_dataset(self, dataset: DataSetFitting):
        """
        判断当前神经网络训练用的数据集和初次训练是用的数据集的归一化参数是否相同，如果不相同，则统一
        """
        if dataset is None or dataset.std_range is None:
            logger.warning(f"训练数据集为空！")
            return True

        if self.std is None and self.mean is None and self.normal_method is None:  # 第一次训练
            logger.debug("首次训练 First train......")
            self.std, self.mean, self.normal_method = dataset.std_range, dataset.mean, dataset.normal_method
            return True

        # 如果self.std不为None，则说明是在原来训练的基础上再次训练
        if dataset.std_range.equals(self.std) and \
                dataset.mean.equals(self.mean) and dataset.normal_method == self.dataset.normal_method:
            return True
        else:
            dataset.set_standard(mean=self.dataset.mean, std_range=self.dataset.std_range,
                                 normal_method=self.dataset.normal_method)
            logger.warning(
                "新数据集与之前训练时使用的数据集的归一化参数不同，这会导致预测结果错误，已自动使用之前训练的归一化参数")
            return False

    def start_train(self, dataset: DataSetFitting, writer=None):
        """
        开始模型训练，多次训练，需要保证dataset数据集中关于归一化的参数（包括归一化方法，数据平均值，标准差等）完全相同，否则会出错。

        :param dataset: DataSetFitting类的对象
        :param writer: tensorboard的记录器
        :return:
        """
        relative_err = []  # 预测结果的相对误差
        if writer is None:
            writer = SummaryWriter(f'./runs')  # 可视化数据存放在这个文件夹
        self.dataset = dataset  # 将传入的dataset设置为神经网络的数据集，该数据集中保存了归一化的参数信息
        self.validate_dataset(dataset)

        writer.add_graph(self, torch.rand([1, self.in_num], device=self.device()))
        train_set, test_set = dataset.split_set()
        train_loader = DataLoader(dataset=train_set, batch_size=self.batch_size, shuffle=True)
        test_loader = DataLoader(dataset=test_set, batch_size=self.batch_size, shuffle=True)
        for t in range(self.epochs):
            logger.debug(f"-------------------------Epoch {t + 1}-------------------------------")
            train_loss, train_err = self.train_loop(train_loader)
            test_loss, (test_relative_err, test_max_err) = self.test_loop(test_loader)
            relative_err.append(test_relative_err)
            if self.classify:
                logger.debug(f"交叉熵损失为：{test_loss:>8f}，分类错误的概率为：{test_relative_err:>8f}")
            else:
                logger.debug(f"平均相对误差为：{test_relative_err:>8f}，最大相对误差为{test_max_err.item():>8f}")
            # ------------------------- 损失曲线图 ---------------------------------
            writer.add_scalar('train/损失', train_loss, t)
            writer.add_scalar('test/平均误差', test_relative_err, t)
            if not self.classify:
                writer.add_scalar('test/最大误差', test_max_err, t)
            # ------------------------- 损失曲线图 ---------------------------------
            # ------------------------- 权重直方图 ---------------------------------
            for i, (name, param) in enumerate(self.named_parameters()):
                if 'bn' not in name:
                    writer.add_histogram(name + "_data", param, t)
                    writer.add_histogram(name + "_grad", param.grad, t)
            # ------------------------- 权重直方图 ---------------------------------
            try:
                if test_relative_err < 0.002 and test_max_err < 0.01:
                    self.max_err = test_max_err
                    self.average_err = test_relative_err
                    break
            except IndexError:  # 当relative_err列表长度小于3时，会报IndexError错误，此处忽略错误继续训练
                pass
            self.max_err = test_max_err
            self.average_err = test_relative_err
            if self.trial is not None:
                self.trial.report(self.average_err, t)
                if self.trial.should_prune():
                    raise optuna.TrialPruned
        logger.debug(f"测试集预测结果的相对误差随epoch的变化为：{[x.item() for x in relative_err]}")
        writer.close()
        logger.debug("\n神经网络训练完成。")
        return self.average_err, self.max_err

    def get_para_type(self, p):
        if isinstance(p, list):
            result = "int"
            for p_ in p:
                if self.get_para_type(p_) == "float":
                    result = "float"
                elif self.get_para_type(p_) == "str":
                    result = "str"
            return result
        else:
            if str(p).strip().isnumeric():  # 小数的isnumeric()方法返回的是false
                return "int"
            else:
                try:
                    _ = eval(p)
                    return "float"
                except NameError:
                    return "str"

    def get_para_of_optuna_str(self, string, name):
        """
        处理设置项中的Optuna优化项，当设置项的值以Optuna开头时，则自动返回Optuna的采样值。如果不是以Optuna开头的变量，则返回原始值

        :param string: 设置项的值
        :param name: 变量名，不影响返回值，但是Optuna工具根据该值识别是哪个变量的采样值，从而给出采样值
        :return:
        """
        result = 0
        if str(string).startswith("optuna"):
            p1 = string.replace("optuna(", "").replace(")", "")
            if "[" not in p1:
                p_list = p1.split(",")
                type1 = self.get_para_type(p_list)
                if type1 == "float":
                    step = None if len(p_list) == 2 else float(p_list[2])
                    result = self.trial.suggest_float(name, float(p_list[0]), float(p_list[1]), step=step)
                elif type1 == "int":
                    step = None if len(p_list) == 2 else int(p_list[2])
                    result = self.trial.suggest_int(name, int(p_list[0]), int(p_list[1]), step=step)
            else:
                p_list = eval(p1)
                result = self.trial.suggest_categorical(name, p_list)
        else:
            result = string
        return result


class OptunaModel:
    def __init__(self, settings=None):
        if settings is None:
            settings = YkDict({})
        self.settings = copy.deepcopy(settings)
        self.settings["networks"] = copy.deepcopy(settings.get_settings("optuna.networks"))
        self.n_trials = int(settings.get_settings("optuna.n_trials") or 10)
        self.device = 'cpu'
        self.mean = None
        self.std = None
        self.x_title = None
        self.y_title = None
        self.train_set = None
        self.test_set = None

    def optimize(self):
        study_name = "model_study20220525"
        study = optuna.create_study(study_name=study_name, direction="minimize",
                                    storage=f'sqlite:///{study_name}.db', load_if_exists=True)
        study.optimize(self.objective, n_trials=self.n_trials)
        trial = study.best_trial
        logger.debug(f"最优模型的损失为{trial.value}")
        logger.debug(f"最优模型的参数为{trial.params}")
        df = study.trials_dataframe()
        print(df)
        self.visualization(study_name)

    @staticmethod
    def visualization(study_name):
        study = optuna.create_study(study_name=study_name, direction="minimize",
                                    storage=f'sqlite:///{study_name}.db', load_if_exists=True)
        optuna.visualization.plot_contour(study).show()  # 参数的等高线图
        optuna.visualization.plot_optimization_history(study).show()  # 优化历史曲线
        optuna.visualization.plot_param_importances(study).show()  # 参数的重要性
        optuna.visualization.plot_intermediate_values(study).show()  # Trails的学习曲线
        optuna.visualization.plot_parallel_coordinate(study).show()  # 高纬度参数的关系，看不懂
        optuna.visualization.plot_slice(study).show()  # 可视化独立参数

    def start_train(self, train_set, test_set):
        self.train_set = train_set
        self.test_set = test_set
        return self.optimize()

    def objective(self, trial):
        model = DataFitterNet(self.settings, trial).to(self.device)
        model.set_standard(self.mean, self.std, self.x_title, self.y_title)
        logger.debug("try model:")
        print(model)
        loss, max_err = model.start_train(self.train_set, self.test_set)
        max_err = max_err.item() if isinstance(max_err, torch.Tensor) else max_err
        loss = loss.item() if isinstance(loss, torch.Tensor) else loss
        trial.set_user_attr('max_err', max_err)
        return loss

    def to(self, device):
        self.device = device
        return self

    def set_standard(self, mean, std, x_title, y_title, *args):
        self.mean = mean
        self.std = std
        self.x_title = x_title
        self.y_title = y_title


class NN:
    def __init__(self, x=None, y=None, dataframe=None, df_x_titles=None, df_y_title=None):
        self.x = x
        self.y = y
        self.dataframe = dataframe
        self.df_x_titles = df_x_titles
        self.df_y_title = df_y_title
        self.settings = {
            "input": df_x_titles,
            "output": df_y_title,
            "networks": {
                "train": {
                    "epochs": 20,
                    "batch_size": 32,
                    "learning_rate": 0.01,
                },
                "loss_fn": {
                    "type": "MSE",

                },
                "optimizer": {
                    "type": "adam",
                },
                "layers": [
                    {"type": "input", "cell_num": "auto"},
                    {"type": "linear", "cell_num": 64, "bias": True},
                    {"type": "relu"},
                    {"type": "linear", "cell_num": 32, "bias": True},
                    {"type": "sigmoid"},
                    {"type": "output", "cell_num": "auto"}
                ],
            },
            "save_path": "default.torch",
            "normalization_method": "z-score"
        }

        self.dataset = DataSetFitting(self.dataframe, self.df_x_titles, self.df_y_title)
        self.dataset.set_standard(normal_method=self.settings.get("normalization_method"))

    def set_batch_size(self, batch_size):
        self.settings["networks"]["train"]["batch_size"] = batch_size

    def set_learning_rate(self, lr):
        self.settings["networks"]["train"]["learning_rate"] = lr

    def set_epochs(self, epochs):
        self.settings["networks"]["train"]["epochs"] = epochs

    def train(self):
        self.net = DataFitterNet(settings1=self.settings)
        self.net.start_train(self.dataset)

    def predict(self, x):
        return self.net.prediction(x)

    def save(self, file):
        self.net.save_yk(file)

    def load(self, file):
        self.net = DataFitterNet.load_yk(file)


def train_model(settings1=None):
    """
    按照settings.yaml文件训练神经网络模型，并保存到settings.yaml指定的位置

    如果settings.networks没有定义具体的网络模型，但是settings.optuna.networks中定义了网络模型的寻优空间参数，则会自动使用
    Optuna库对模型超参数进行寻优，并将寻优结果得到的最有网络模型定义写入settings.yml文件中。再次调用该方法，则会使用寻优得到的
    网络模型架构进行数据拟合，将模型权值信息写入settings.output.save_path指定的路径中。

    如果settings.networks定义了具体的网络模型，则该方法只对定义的模型进行训练，将训练后的权值信息写入settings.output.save_path
    指定的路径中。

    如果该方法同时训练多个单输出模型，则模型架构的优化只会进行一次，即第一个模型架构的优化结果会应用于后续的多个单输出数据拟合任务。

    :return:
    """
    device = "cuda" if torch.cuda.is_available() else "cpu"
    logger.info(f"使用{device}进行计算")
    if settings1 is None:
        settings1 = get_settings()

    if settings1.get("networks") is not None:  # 如果存在networks的定义就直接使用该定义
        optimize_model = False
    elif settings1.get("optuna") is not None:  # 如果存在optuna的设置，就使用该设置获取最有网络模型定义
        optimize_model = True
    else:
        logger.error("未找到模型设置信息，请确认！")
        return
    _settings_ = copy.deepcopy(settings1)
    output = _settings_.get("output") or {}
    if output.get("type") == "single_output":
        paras = copy.deepcopy(output.get("para") or {})
        for i, para in enumerate(paras):
            output["para"] = [para]
            output["save_path"] = settings1.get_settings("output.save_path")[i]
            model = OptunaModel(_settings_).to(device) if optimize_model else DataFitterNet(_settings_).to(device)
            logger.debug(model)
            dataset_all = DataSetFitting(_settings_)  # 获取所有的数据集
            model.set_standard(dataset_all.mean, dataset_all.std_range, dataset_all.x_title, dataset_all.y_title,
                               dataset_all.normal_method)  # 将数据集的数据归一化信息添加到模型中，方便以后使用
            part1 = float(_settings_.get_settings("dataset.data_training.proportion"))
            part2 = float(_settings_.get_settings("dataset.data_test.proportion"))
            train_set, test_set = dataset_all.split_set(part1, part2)  # 均已是归一化后的数据
            model.start_train(train_set, test_set)
            if not optimize_model:
                model.save_yk(output["save_path"])
            else:
                model.update_settings_file("settings.yml")
                train_model()


def re_train(data_file, settings1=None):
    """
    在现有模型基础上再次训练，该方法对数据的归一化方法必须与初次训练时一致，因此本方法会忽略settings1中的数据归一化相关设置，而自动采用原模型中
    的数据归一化方法

    :param data_file: 数据文件,csv文件
    :param settings1: 可以直接传入settings字典，则该方法不会再从外部yaml文件读取设置信息
    :return:
    """
    device = "cuda" if torch.cuda.is_available() else "cpu"
    logger.info(f"使用{device}再次训练")
    if settings1 is None:
        settings1 = get_settings()
    _settings_ = copy.deepcopy(settings1)
    if isinstance(data_file, list):
        _data_file = []
        for file in data_file:
            _data_file.append({"name": file, "type": "csv"})
        _settings_["dataset"]["data_file"] = _data_file
    else:
        _settings_["dataset"]["data_file"] = [{"name": data_file, "type": "csv"}]
    output = _settings_.get("output") or {}
    if output.get("type") == "single_output":
        paras = copy.deepcopy(output.get("para") or {})
        for i, para in enumerate(paras):
            output["para"] = [para]
            output["save_path"] = settings1.get_settings("output.save_path")[i]
            model = DataFitterNet.load_yk(output["save_path"])
            model.epochs = _settings_.get_settings("networks.train.epochs")  # 再次训练时的epochs和初次训练时的一般不相同
            logger.debug(model)
            dataset_all = DataSetFitting(_settings_)  # 获取所有的数据集
            if _settings_.get_settings("dataset.normalization") != model.normal_method:
                logger.warning(f"再训练时数据归一化方法({_settings_.get_settings('dataset.normalization')})与"
                               f"模型初次训练时({model.normal_method})不一致，这可能导致预测结果错误")
                logger.warning(f"已强制切换为模型初次训练时使用的数据归一化方法！({model.normal_method})")
            dataset_all.set_standard(model.mean, model.std, model.normal_method)
            part1 = 0.95
            part2 = 0.05
            train_set, test_set = dataset_all.split_set(part1, part2)  # 均已是归一化后的数据
            model.start_train(train_set, test_set)

            model.save_yk(output["save_path"])


def filter_err_condition(model, data_file=None):
    """
    将model在整个数据集上进行测试，过滤小偏差的点，删选出偏差较大的点，一般后续对偏差较大的点继续学习

    :param model:
    :param data_file:
    :return:
    """
    settings = get_settings()
    if data_file is not None:
        settings["dataset"]["data_file"] = [{"name": data_file, "type": "csv"}]
    _settings_ = copy.deepcopy(settings)
    dataset_all = DataSetFitting(_settings_)
    dataset_all.set_standard(model.mean, model.std_range)


def setup_seed(seed=10):
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)


def test():
    model = DataFitterNet.load_yk(r"D:\lengduan\data\model_p2.dat")
    args = {
        "power": 400,
        "flow_heat": 0,
        "p_env": 98,
        "t_env": 30,
        "humid": 0.60,
        "p_gas": 3.8,
        "t_gas": 18,
        "flow_fgh": 40,
        "flow_tca": 100,
        "flow_oh": 3,
        "flow_rh": 0
    }
    x = [v for k, v in args.items()]
    pump, fun = 2, 3
    x.extend([pump, fun])
    p = model.prediction(x)
    print(p)


def deal_settings_path(settings, **kwargs):
    """
    当settings.yaml文件中使用了带变量的路径时，如"E:/热工院//retrain{unit}_0.csv",
    本方法将变量{unit}替换为指定的变量。

    :param settings:
    :return:
    """
    files = []

    for data_file in settings.get_settings("dataset.data_file"):
        pth = data_file["name"].replace(r"{unit}", "1")
        type1 = data_file["type"]
        files.append({"name": pth, "type": type1})
    settings["dataset"]["data_file"] = files
    return settings


if __name__ == "__main__":
    # setup_seed()
    # OptunaModel(None).visualization("model_study20220525")
    # settings1 = get_settings()
    # deal_settings_path(settings1)
    # train_model(settings1)
    # # settings["output"]["save_path"] = [r"D:\lengduan\data\model_p1.dat"]
    # # settings["output"]["para"] = ["背压"]
    # re_train(data_file=[
    #     r"D:\lengduan\data\retrain1_0.csv",
    #     r"D:\lengduan\data\retrain1_1.csv",
    #     r"D:\lengduan\data\retrain1_7.csv",
    #     r"D:\lengduan\data\retrain1_9.csv",
    #     r"D:\lengduan\data\retrain1_10.csv",
    #     r"D:\lengduan\data\retrain1_11.csv",
    # ], settings1=settings_1)
    # re_train(data_file=[
    #     r"D:\lengduan\data\retrain2_0.csv",
    #     r"D:\lengduan\data\retrain2_1.csv",
    #     r"D:\lengduan\data\retrain2_7.csv",
    #     r"D:\lengduan\data\retrain2_8.csv",
    #     r"D:\lengduan\data\retrain2_9.csv",
    #     r"D:\lengduan\data\retrain2_10.csv",
    #     r"D:\lengduan\data\retrain2_11.csv",
    # ], settings1=settings_1)

    import torch
    import torch.autograd

    x = torch.tensor([2.0], requires_grad=True)
    y = x ** 3

    y.backward(retain_graph=True)
    y.backward(retain_graph=True)
    y.backward()  # 反向传播,求解导数，即y对x求导
    print("x.grad = ", x.grad)



"""
绘制热力系统图

"""
import cocos
import pyglet
from cocos.actions import *
from cocos.director import director
from cocos.scene import Scene
from cocos.cocosnode import CocosNode
from pyglet.canvas import Canvas

from yangke.performance.da.Components import Boiler


class KeyDisplay(cocos.layer.Layer):
    is_event_handler = True

    def __init__(self):
        super(KeyDisplay, self).__init__()
        self.text = cocos.text.Label("", x=100, y=280)
        self.key_pressed = set()
        self.update_text()
        self.add(self.text)

    def update_text(self):
        key_names = [pyglet.window.key.symbol_string(k) for k in self.key_pressed]
        text = 'Keys: ' + ','.join(key_names)
        self.text.element.text = text

    def on_key_press(self, key, modifiers):
        self.key_pressed.add(key)
        self.update_text()

    def on_key_release(self, key, modifiers):
        self.key_pressed.remove(key)
        self.update_text()


class MouseDisplay(cocos.layer.Layer):
    is_event_handler = True

    def __init__(self):
        super(MouseDisplay, self).__init__()
        self.posx = 100
        self.posy = 240
        self.text = cocos.text.Label('No mouse event yet',
                                     font_size=18,
                                     x=self.posx, y=self.posy)
        self.add(self.text)

    def update_text(self, x, y):
        text = f"Mouse @ {x},{y}"
        self.text.element.text = text
        self.text.element.x = self.posx
        self.text.element.y = self.posy

    def on_mouse_motion(self, x, y, dx, dy):
        self.update_text(x, y)

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        self.update_text(x, y)

    def on_mouse_press(self, x, y, buttons, modifiers):
        self.posx, self.posy = director.get_virtual_coordinates(x, y)
        self.update_text(x, y)


class MyApp(cocos.layer.ColorLayer):
    def __init__(self):
        super(MyApp, self).__init__(255, 255, 255, 255)
        # super(MyApp, self).__init__()

        # label = cocos.text.Label('Hello, World!',
        #                          font_name='Times New Roman',
        #                          font_size=32,
        #                          anchor_x='center', anchor_y='center',
        #                          position=(320, 240))
        # self.add(label)
        self.add(Boiler(), z=1)

        # sprite = cocos.sprite.Sprite("grossini.png", position=(320, 240), scale=1)
        # self.add(sprite, z=1)
        # scale = ScaleBy(3, duration=2)  # 两秒内缩放三次
        # label.do(Repeat(scale + Reverse(scale)))  # 将缩放动作添加到label标签上


if __name__ == "__main__":
    director.init(resizable=True, width=1400, height=800)
    # app_layer = MyApp()
    # app_layer.do(RotateBy(360, duration=10))  # 让整个画面旋转
    # director.run(Scene(app_layer))
    director.打开技术报告页面(Scene(MyApp()))



"""
向Word页面中添加水印的方法：
1.从已经编号的PDF报告中，抠出背景透明的水印.png图片
    1.1 使用WPS打开PDF文件，提取水印与文字重叠最少的一页至新的PDF文件中。
    1.2 使用GIMP2.0打开提取的PDF，分辨率设置为300。另存为“水印.jpg”文件
    1.3 使用GIMP2.0打开“水印.jpg"文件，删除除水印外的其他文字或图标等页面元素，然后选择水印文字区，以纯黑色填充水印选取，然后导出为”水印.png"文件。
    1.4 使用GIMP2.0打开“水印.png"文件，添加透明通道后，使用颜色选择工具扣出水印文字，并删除背景，使背景为透明，并保存。
    注意：以上操作过程中，不改变图片大小，不进行裁剪。
2.将水印图片加入修改后的WORD页面中生成对应页面的PDF文件。
    2.1 将修改后的WORD页面转换为PDF文件，命名为1.pdf。
    2.2 使用WPS打开1.pdf，选择【编辑】-【水印】-【添加】/【更新水印】（第一次为添加，后续为更新），打开水印编辑界面。
    2.3 在水印编辑界面中，选择水印来源为”水印.png"，旋转为"水平"，不透明度为"18%"，相对页面比例为"100%"，垂直对齐为0，水平对齐为"居中2mm",
        然后确定即可。
    2.4
3. 将修改后的PDF页面替换至原PDF文件中。
    3.1 使用WPS打开原PDF文件，选择【页面】-【替换页面】，在弹出窗口中选择第2步生成的PDF即可。
"""


# encoding=utf-8
from direct.showbase.ShowBase import ShowBase
from panda3d.core import WindowProperties, TextNode, load_prc_file_data, ConfigVariableString
from direct.gui.DirectGui import *

"""
panda3d中的CamelCase和camel_case格式的函数名是同样功能的，只是一个是C++命名规则，一个是python命名规则
"""

# text-encoding
load_prc_file_data("", "text-default-font /c/Windows/Fonts/simsun.ttc")  # 使panda3D支持中文


class Game(ShowBase):
    def __init__(self):
        super(Game, self).__init__()
        self.text1, self.text2, self.game_over_screen, self.final_score_label = None, None, None, None
        self.init2d()
        # self.scene = self.loader.loadModel("models/environment")
        self.scene = self.loader.loadModel("models/海马/haima")
        self.scene.reparentTo(self.render)
        self.scene.setScale(0.25, 0.25, 0.25)
        self.scene.setPos(-8, 42, 0)
        # self.disable_mouse()

    def init2d(self):
        properties = WindowProperties()
        properties.setSize(1400, 750)
        properties.setTitle("性能监测及运行优化系统")
        self.win.requestProperties(properties)
        bk_text = "DirectOptionMenu Demo"
        self.text1 = OnscreenText(text=bk_text, pos=(0.85, 0.85), scale=0.07,
                                  fg=(1, 0.5, 0.5, 1), align=TextNode.ACenter,
                                  mayChange=1)
        output = "item "
        self.text2 = OnscreenText(text=output, pos=(0.95, -0.95), scale=0.07,
                                  fg=(1, 0.5, 0.5, 1), align=TextNode.ACenter,
                                  mayChange=1)
        # menu = DirectOptionMenu(text="options", scale=0.1, command=self.item_sel,
        #                         items=["item1", "item2", "item3"], initialitem=2,
        #                         highlightColor=(0.0, 0.65, 0.65, 1), pos=(0, 0, 0))

        self.game_over_screen = DirectDialog(frameSize=(-0.7, 0.7, -0.7, 0.7), fadeScreen=0.4, relief=DGG.FLAT,
                                             frameTexture="UI/stoneFrame.jpg")
        label = DirectLabel(text="Game Over!", parent=self.game_over_screen, scale=0.1, pos=(0, 0, 0.2))

        self.final_score_label = DirectLabel(text="", parent=self.game_over_screen, scale=0.07, pos=(0, 0, 0))
        btn = DirectButton(text="重新开始", command=self.start_game, pos=(-0.3, 0, -0.2), parent=self.game_over_screen,
                           scale=0.07)
        DirectButton(text="退出", command=self.quit, pos=(0.3, 0, -0.2), parent=self.game_over_screen, scale=0.07)
        self.game_over_screen.hide()

    def item_sel(self, arg):
        output = "选择了选项：" + arg
        self.text2.setText(output)

    def start_game(self):
        ...

    def quit(self):
        ...


game = Game()
game.run()



from twisted.internet.protocol import Protocol, ClientFactory
from twisted.internet import reactor
import sys
from datetime import datetime


class Echo(Protocol):
    def __init__(self, factory):
        self.factory = factory

    def connectionMade(self):
        self.factory.num
        print("Connected to the server!")

    def dataReceived(self, data):
        print("got messages: ", data.decode())
        reactor.callLater(5, self.say_hello)

    def connectionLost(self, reason=connectionDone):
        print("Disconnected from the server!")

    def say_hello(self):
        if self.transport.connected:
            self.transport.write(u"hello, I'm {} {}".format(sys.argv[1], datetime.now()))



import pygame
from pygame.sprite import Sprite


class GridSprite(Sprite):
    """
    网格线
    """

    def __init__(self, x_list, y_list, width=10, height=10, color=(0, 0, 0), line_width=1):
        """
        pygame的网格线部件
        :param x_list: 绘制的垂直于x轴的网格线的位置列表
        :param y_list: 绘制的垂直于y轴的网格线的位置列表
        :param width: 绘图界面的宽，也就是画布的宽
        :param height: 绘图界面的高，也就是画布的高
        :param color: 网格线的颜色
        :param line_width: 网格线的宽度
        """
        super(GridSprite, self).__init__()
        self.x_list = x_list
        self.y_list = y_list
        self.color = color
        self.line_width = line_width
        self.image = pygame.Surface([width, height]).convert_alpha()
        self.image.fill((255, 255, 255, 0))
        # self.image.set_alpha(128)  # 0全透明，看不到当前层，255完全不透明，看不到被遮挡的层
        self.rect = self.image.get_rect()
        self.draw()

    def update(self, x_list, y_list):
        self.x_list = x_list
        self.y_list = y_list
        self.draw()

    def draw(self, screen=None):
        """
        精灵的绘制会遮挡self.rect下的其他内容，没找到解决方法

        :param screen
        :return:
        """
        canvas = self.image
        canvas_width = canvas.get_width()
        canvas_height = canvas.get_height()
        if self.x_list is not None:
            for x in self.x_list:
                pygame.draw.line(canvas, color=self.color, start_pos=(x, 0), end_pos=(x, canvas_height), width=1)
        if self.y_list is not None:
            for y in self.y_list:
                pygame.draw.line(canvas, color=self.color, start_pos=(0, y), end_pos=(canvas_width, y), width=1)
        if screen is not None:
            screen.blit(self.image, (0, 0))


class YkPygameSprite(Sprite):
    def __init__(self):
        super(YkPygameSprite, self).__init__()
        self.position = (0, 0)
        self.connect_points = {}  # 记录图标上的锚点
        self.connect_points_dir = {}  # 记录锚点的方向
        self.connect_points_lines = {}  # 记录锚点与其他图标连接的线信息
        self.tag = None

    def move_to(self, x, y):
        """
        移动到位置(x,y)

        :param x:
        :param y:
        :return:
        """
        self.position = x, y
        self.scale_by(1)

    def scale_by(self, scale):
        w, h = self.image.get_width(), self.image.get_height()
        w_new, h_new = int(w * scale), int(h * scale)
        self.image = pygame.transform.scale(self.image, (w_new, h_new))
        self.rect = pygame.rect.Rect(self.position[0] - w_new // 2, self.position[1] - h_new // 2, w_new, h_new)

    def get_absolute_connect_points(self):
        """
        获取图标连接点的绝对坐标
        :return:
        """
        x, y = self.position
        connect_points = {}
        for name, pos in self.connect_points.items():
            if pos is not None:
                connect_points.update({name: (pos[0] + x, pos[1] + y)})
            else:
                connect_points.update({name: None})
        return connect_points


class BoilerDiagram(YkPygameSprite):
    """
    锅炉
    """

    def __init__(self, position=(0, 0), color=(0, 0, 0), line_width=2):
        super(BoilerDiagram, self).__init__()
        self.is_event_handler = False
        self.position = position  # 锅炉位置坐标
        self.line_width = line_width  # 线宽
        self.color = color  # 线条颜色
        # self.rect1 = [0, 0, 80, 100]

        self.image = pygame.Surface([80 + line_width, 120 + line_width]).convert_alpha()
        self.image.fill((255, 255, 255, 0))

        self.draw()
        width = self.image.get_width()
        height = self.image.get_height()
        self.rect = pygame.rect.Rect(position[0] - width // 2, position[1] - height // 2, width, height)

        self.connect_points = {"主给水连接点": (-10, height//2), "主蒸汽连接点": (-10, -height//2),
                               "冷再连接点": None, "热再连接点": None}
        self.connect_points_dir = {"主给水连接点": "down", "主蒸汽连接点": "up",
                                   "冷再连接点": "down", "热再连接点": "up"}

    def draw(self):
        """
        绘制锅炉图标

        :return:
        """
        canvas = self.image
        pygame.draw.rect(canvas, width=self.line_width, color=self.color, rect=[0, 10, 80, 100])
        pygame.draw.lines(canvas, width=self.line_width, color=self.color, closed=False,
                          points=[(30, 0), (30, 20), (30, 100), (30, 120)])
        pygame.draw.lines(canvas, width=self.line_width, color=self.color, closed=False,
                          points=[(50, 0), (50, 20), (50, 100), (50, 120)])


class TurbineHigh(YkPygameSprite):
    def __init__(self, position=(0, 0), line_width=2, color=(0, 0, 0)):
        super(TurbineHigh, self).__init__()
        self.position = position
        self.line_width = line_width
        self.color = color

        self.center_line = (0, 100)
        self.image = pygame.Surface([120, 140]).convert_alpha()
        self.image.fill((0, 0, 255, 50))  # 第四个参数默认是255
        width = self.image.get_width()
        height = self.image.get_height()
        self.rect = pygame.rect.Rect(position[0] - width // 2, position[1] - height // 2, width, height)
        self.draw()

        self.connect_points = {"主蒸汽连接点": (118, 20), "冷再连接点": (0, 138), "1抽连接点": None}
        self.connect_points_dir = {"主蒸汽连接点": "up", "冷再连接点": "down", "1抽连接点": "down"}

    def draw(self):
        """
        绘制高压缸图标

        :return:
        """
        canvas = self.image
        pygame.draw.lines(canvas, width=self.line_width, color=self.color, closed=True,
                          points=[(0, 0), (118, 20), (118, 118), (0, 138)])


class TurbineMiddle(YkPygameSprite):
    def __init__(self, position=(0, 0), line_width=2, color=(0, 0, 0)):
        super(TurbineMiddle, self).__init__()
        self.position = position
        self.line_width = line_width
        self.color = color

        self.center_line = (0, 100)
        self.image = pygame.Surface([160, 140]).convert_alpha()
        self.image.fill((255, 255, 250, 0))
        width = self.image.get_width()
        height = self.image.get_height()
        self.rect = pygame.rect.Rect(position[0] - width // 2, position[1] - height // 2, width, height)
        self.draw()

        self.connect_points = {"热再连接点": (118, 20), "3抽连接点": (0, 138), "4抽连接点": None,
                               "中排连接点": (158, 0)}
        self.connect_points_dir = {"热再连接点": "up", "3抽连接点": "down", "4抽连接点": "down",
                                   "中排连接点": "down"}

    def draw(self):
        """
        绘制高压缸图标

        :return:
        """
        canvas = self.image
        pygame.draw.lines(canvas, width=self.line_width, color=self.color, closed=True,
                          points=[(0, 20), (158, 0), (158, 138), (0, 118)])


class Heater(YkPygameSprite):
    def __init__(self, position=(0, 0), line_width=2, color=(0, 0, 0)):
        super(Heater, self).__init__()
        self.position = position
        self.line_width = line_width
        self.color = color

        self.image = pygame.Surface([80, 40]).convert_alpha()
        self.image.fill((255, 255, 255, 0))
        width = self.image.get_width()
        height = self.image.get_height()
        self.rect = pygame.rect.Rect(position[0] - width // 2, position[1] - height // 2, width, height)
        self.draw()

        self.connect_points = {"进水连接点": (78, 20), "进汽连接点": (20, 0), "出水连接点": (0, 20),
                               "疏水出口连接点": (70, 30), "上级连排疏水连接点": None}
        self.connect_points_dir = {"进水连接点": "down", "进汽连接点": "up", "出水连接点": "left",
                                   "疏水出口连接点": "down", "上级连排疏水连接点": "down"}

    def draw(self):
        """
        绘制高压缸图标

        :return:
        """
        canvas = self.image
        pygame.draw.lines(canvas, width=self.line_width, color=self.color, closed=True,
                          points=[(0, 0), (78, 0), (78, 38), (0, 38)])
        pygame.draw.line(canvas, color=self.color, start_pos=(58, 0), end_pos=(58, 40), width=self.line_width)



import traceback

import numpy as np
import pandas as pd
# import pickle
import dill as pickle
import re
import os
import time
import datetime

import pandas.errors

from yangke.base import add_sep_to_csv, get_encoding_of_file
from yangke.common.config import logger


def get_last_modified_time(file: str):
    last_change_time = os.stat(file).st_mtime
    last_change_time_str = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(last_change_time))
    last_change_time = datetime.datetime.strptime(last_change_time_str, "%Y-%m-%d %H:%M:%S")
    return last_change_time


class re_common:
    """
    常见正则表达式的操作

    Example:
        1. 需要删除括号及括号中的内容，则使用以下语句：

        re = fo.re_common("前者多指对文章（书籍）中某一部分，但为了防止冗杂而把它放在段落之外（文末或页边）")
        result = re.del_char_in_brackets("（", "）")


    """

    def __init__(self, content):
        """
        用需要处理的目标字符串初始化re_common类对象

        :param content: 需要处理的目标字符串
        """
        self.content = content

    def del_char_in_brackets(self, left="(", right=")"):
        """
        如果content中包含括号，则删除括号及括号中的内容。

        :param left: 左括号的字符，如【、{、（、<、<-、《等
        :param right: 右括号的字符
        :return:
        """
        content_ = self.content
        result = re.match(f".*{left}.+{right}.*", content_)  # 判断content中是否存在（）
        while result:
            temp_list = list(re.findall(f"(.*){left}.+{right}(.*)", content_)[0])  # 正则中()中的内容会保留
            temp_list = [item for item in temp_list if item != ""]  # 删掉空字符串
            content_ = "".join(temp_list)  # 拼接起来
            result = re.match(f".*{left}.+{right}.*", content_)  # 判断content中是否存在（），存在就继续去除
        self.content = content_
        return self.content


def read_data(file, encoding="utf8"):
    """
    读取股票数据的csv文件内容到
    :param file:
    :param encoding:
    :return:
    """
    f = open(file)
    df = pd.read_csv(f, encoding=encoding)
    data_local = df.iloc[:, 1:6]
    return data_local.values


def write_line(file: str, line: str, encoding="utf8", append=False):
    """
    将字符串line内容写入文件，如果文件不存在则创建，如果文件存在，默认覆盖原文件，可以通过设置append=True实现追加文本
    """
    mode = 'a' if append else 'w'
    file = os.path.abspath(file)
    if not os.path.exists(os.path.dirname(file)):
        try:
            os.makedirs(os.path.dirname(file))
        except PermissionError:
            logger.error(f"权限不足，无法写入文件{file}")
    with open(file, encoding=encoding, mode=mode) as f:
        f.write(line)


def write_lines(file: str, lines: list, encoding="utf8", append=False):
    """
    将字符串列表写入文件，每个列表项为单独一行
    """
    import os
    mode = 'a' if append else 'w'
    file = os.path.abspath(file)
    if not os.path.exists(os.path.dirname(file)):
        try:
            os.makedirs(os.path.dirname(file))
        except PermissionError:
            logger.error(f"权限不足，无法写入文件{file}")
    # lines = os.linesep.join(lines)  # 列表项之间添加换行符，os.linesep会换两行
    lines = "\n".join(lines)
    with open(file, encoding=encoding, mode=mode) as f:
        f.writelines(lines)


def read_lines(file: str, dropna=False) -> list:
    """
    读取文件中的所有行至列表，可以自动识别文本的编码格式

    :param file:
    :param dropna: 是否丢弃空行，默认不丢弃
    :return:
    """
    encoding = get_encoding_of_file(file)
    with open(file, encoding=encoding, mode='r') as f:
        lines = f.readlines()

    result = []
    for line in lines:
        if line.strip():
            line = line.rstrip("\n").rstrip("\r")
            result.append(line)  # 删除行右侧的换行符

    return result


def read_points(file: str, split=',', return_type='[x][y][z]'):
    """
    从txt文件中读取点坐标

    :param file:
    :param split:
    :param return_type: '[x][y][z]'则返回x,y,z三个列表，如果是'[xyz]'则返回[x,y,z]形式的点坐标列表
    :return:
    """
    points = []
    x, y, z = [], [], []
    with open(file, mode='r') as f:
        for line in f.readlines():
            coor = line.split(split)
            px = float(coor[0])
            py = float(coor[1])
            pz = float(coor[2])
            points.append([px, py, pz])
            x.append(px)
            y.append(py)
            z.append(pz)

    if return_type == '[xyz]':
        return points
    else:
        return x, y, z


def write_as_pickle(file: str, obj: object):
    """
    保存任意对象到硬盘文件，obj是函数对象时，保存的是函数的名称和地址，无法在应用重启后加载原函数。
    如果需要保存函数，请使用write_func(file: str, func: object)

    :param file:
    :param obj:
    :return:
    """
    try:
        file = os.path.abspath(file)
        if not os.path.exists(os.path.dirname(file)):
            os.makedirs(os.path.dirname(file))
        with open(file, 'wb') as f:
            pickle.dump(obj, f)
    except OSError:
        traceback.print_exc()
        logger.error(f"保存对象时发生错误，保存路径为{file}，对象为{obj}")
    except:
        traceback.print_exc()
        logger.error(f"保存对象时发生错误，保存路径为{file}，对象为{obj}")


def read_from_pickle(file: str):
    """
    从硬盘文件加载pickle对象，obj是函数对象时，请使用read_func(file: str, func: object)，文件不存在时返回None

    :param file: 硬盘文件
    :return:
    """
    if file is None:
        return None
    if not os.path.exists(file):  # 文件不存在，返回None
        return None
    obj = None
    try:
        with open(file, 'rb') as f:
            obj = pickle.load(f)
    except:
        traceback.print_exc()
        logger.error(f"加载文件{f}时发生错误！")
    return obj


def write_func(file: str, func: object):
    """
    保存python函数或方法到硬盘，以便应用重启后直接加载

    :param file: 保存到的文件名
    :param func: 需要保存的函数名
    :return:
    """
    import dill
    write_as_pickle(file, dill.dumps(func))


def read_func(file: str):
    """
    从硬盘文件中加载函数对象

    :param file:
    :return:
    """
    import dill
    return dill.loads(read_from_pickle(file))


def read_from_yaml(file: str, encoding="utf8"):
    import yaml
    if not os.path.exists(file):  # 文件不存在，返回空字典
        return {}
    with open(file, 'r', encoding=encoding) as f:
        content = f.read()
    """
    Loader的几种加载方式 
    BaseLoader--仅加载最基本的YAML 
    SafeLoader--安全地加载YAML语言的子集。建议用于加载不受信任的输入。 
    FullLoader--加载完整的YAML语言。避免任意代码执行。这是当前（PyYAML 5.1）默认加载器调用 
            yaml.load(input)（发出警告后）。
    UnsafeLoader--（也称为Loader向后兼容性）原始的Loader代码，可以通过不受信任的数据输入轻松利用。"""
    obj = yaml.load(content, Loader=yaml.FullLoader)
    return obj


def write_as_yaml(file, obj):
    import yaml
    with open(file, 'w') as f:
        f.write(yaml.dump(obj, allow_unicode=True))


def write_excel(file, dataframe_dict: dict, index=False):
    with pd.ExcelWriter(file) as writer:
        for sheet, df in dataframe_dict.items():
            df.to_excel(writer, sheet_name=sheet, index=index)
    return


def read_excel(file, need_merge=False, index_col=True):
    """
    将excel文件中的所有sheet读取到dataframe_dict中

    :param need_merge: 是否合并各个sheet中的数据，默认不合并，且只有各个sheet中数据格式相同时才能合并
    :param file:
    :param index_col: 是否需要索引，也可以是数字
    :return:
    """
    if index_col:
        index_col = 0
    file = pd.ExcelFile(file)
    sheets = file.sheet_names
    if not need_merge:
        dataframe_dict = {}
        for sheet in sheets:
            df = pd.read_excel(file, sheet_name=sheet, index_col=index_col)
            dataframe_dict.update({sheet: df})
        return dataframe_dict
    else:
        from yangke.base import merge_two_dataframes, merge_dataframes_simple
        res = None
        for sheet in sheets:
            if res is None:
                res = pd.read_excel(file, sheet_name=sheet, index_col=index_col)
            else:
                res = merge_dataframes_simple([res, pd.read_excel(file, sheet_name=sheet, index_col=index_col)])

        return res


def read_csv_assist(file, sep=",", encoding="utf-8"):
    data = []
    with open(file, 'r', encoding=encoding) as f_input:
        for line in f_input:
            data.append(list(line.strip().split(sep)))
    return pd.DataFrame(data)


def read_csv_ex(file, sep=None, header="infer", skiprows=None, on_bad_lines='skip', nrows=None, index_col=None,
                low_memory=None, na_values=" ") -> pd.DataFrame:
    """
    pandas增强版的read_csv()方法，可以自动匹配任何文件编码，自动补全数据列不全的行

    :param skiprows:
    :param low_memory:
    :param index_col:
    :param nrows:
    :param file:
    :param sep:
    :param header:
    :param on_bad_lines:
    :param na_values: 设置na_values=" "，则pandas在读取到" "时，会替换为数值类型的np.nan，如果不设置，则缺失数据的列会被视为object，导致无法进行数学计算
    :return:
    """
    encoding_csv = get_encoding_of_file(file)
    to_file1 = os.path.join(os.path.dirname(file), f"{os.path.basename(file)[:-4]}_temp_修正.csv")
    if file[-3:].lower() == "txt" and sep is None:  # txt文件默认以空格作为分隔符
        sep = "\s+"
    to_file1 = add_sep_to_csv(file, sep, to_file=to_file1, encoding=encoding_csv)
    if file[-3:].lower() == "txt" and sep == "\s+":  # 如果该条件满足，则add_sep_to_csv已经将分隔符替换为"\t"
        sep = "\t"
    if sep is None:
        sep = ","
    try:
        data = pd.read_csv(to_file1, sep=sep, header=header, on_bad_lines=on_bad_lines,
                           encoding=encoding_csv, skiprows=skiprows, nrows=nrows, index_col=index_col,
                           low_memory=low_memory, na_values=na_values)
    except pandas.errors.ParserError:  # 有可能是双引号导致的错误，删除双引号重新尝试一次
        to_file1 = add_sep_to_csv(file, sep, to_file=to_file1, encoding=encoding_csv, delete_quote=True)
        data = pd.read_csv(to_file1, sep=sep, header=header, error_bad_lines=on_bad_lines,
                           encoding=encoding_csv, skiprows=skiprows, nrows=nrows, index_col=index_col,
                           low_memory=low_memory, na_values=na_values)
    if os.path.exists(to_file1):
        os.remove(to_file1)
    return data


def pd_set_na(df, conditions, expand=False, axis=0):
    """
    将dataframe中满足指定条件的行的元素置为np.nan
    示例：
    pd_set_na(df, {"大气压力": {"<": 0.03}}, expand=True, axis=0)
    pd_set_na(df, {"大气压力": {"<": 0.03, ">": "0.2"}}, expand=True, axis=0)
    pd_set_na(df, {"大气压力": "大气压力<0.03 and 环境温度<0.2", expand=True, axis=0)
    pd_set_na(df, {"大气压力": "大气压力<0.03 or 大气压力>0.2", expand=True, axis=0)

    :param df:
    :param conditions:
    :param expand:
    :param axis:
    :return:
    """
    for title, cond in conditions.items():
        cond: dict = cond
        for op, val in cond.items():
            if op == "<":
                if expand:
                    for col in df.columns:
                        if col != title:
                            df.loc[df[title] < val, col] = np.nan
                    df.loc[df[title] < val, title] = np.nan  # 最后处理title列对应的行，防止数据改变导致条件失效
                else:
                    df.loc[df[title] < val, title] = np.nan
            elif op == ">":
                if expand:
                    for col in df.columns:
                        if col != title:
                            df.loc[df[title] > val, col] = np.nan
                    df.loc[df[title] > val, title] = np.nan  # 最后处理title列对应的行，防止数据改变导致条件失效
                else:
                    df.loc[df[title] > val, title] = np.nan


def fill_excel_cell(file, sheet: int | str = 0, row=1, col: int | str = 1, cell=None, value=None, visible=True,
                    close=True, res=None):
    """
    在excel文件中动态填入参数，并自动计算一次excel。使用示例：
    fill_excel_cell('test.xlsx", "Sheet1", row=1, col="C", value="hello world")

    入股需要多次填充同一个excel文件，为了避免反复打开关闭文件，可以设置close和res参数。示例如下：
    res = fill_excel_cell('test.xlsx", "Sheet1", cell="A1", value="hello world", close=False)  # 第一次close=False
    res = fill_excel_cell('test.xlsx", "Sheet1", cell="A2", value="value A2", close=False, res=res) #第二次res传入第一次的结果
    ...
    res = fill_excel_cell('test.xlsx", "Sheet1", cell="A3", value="value A3", close=True, res=res) # 最后一次close=True

    :param file: 需要填入参数的Excel文件路径
    :param sheet: Excel中的sheet名或索引
    :param row:
    :param col:
    :param cell: 以"A1"形式标记的单元格坐标
    :param value:
    :param visible:
    :param close: 填充后是否关闭excel文件
    :param res: 如果上次未关闭excel文件，则通过该参数传入上次打开的文件信息
    :return:
    """
    if res is None:
        import xlwings as xw
        if not os.path.exists(os.path.abspath(file)):
            logger.error(f"操作的excel文件不存在 : {os.path.abspath(file)}")

        app = xw.App(visible=visible, add_book=False)
        app.display_alerts = False
        app.screen_updating = visible
        # app.calculation = "manual"
        wb = app.books.open(file)
    else:
        app, wb = res
    sht = wb.sheets[sheet]
    if cell is not None:
        sht.range(cell).value = value
    else:
        sht.range(f"{col}{row}").value = value

    if close:
        app.calculate()
        wb.save()
        app.quit()
    else:
        return app, wb


if __name__ == "__main__":
    fill_excel_cell(file=r'C:\Users\YangKe\Desktop\碳排放量计算.xlsx', sheet='附表 C.3 化石燃料燃烧排放表', row=3,
                    value=23,
                    col='E')



"""
1.安装office
2.pip install xlwings
3.xlwings addin install
4.在excel中xlwings下设置python解释器的路径
5.有必要的在excel中xlwings下设置模块的路径和需要引入的模块名

"""
import os
from yangke.base import find_file


def auto_set():
    """
    将python路径设置
    自动设置xlwings
    :return:
    """
    user_dir = os.path.expanduser("~")
    settings_file = find_file("*/xlwings.conf", user_dir)
    if len(settings_file) == 0:
        print("未找到xlwings的设置文件")
        return False
    setting_file = settings_file[0]



from yangke.performance.tools.natural_gas import NaturalGas


def c_combustion_gas(flow_gas=10000, natural_gas=None):
    """
    根据天然气流量计算实时碳排放，参考《GB/T 32151》，如果不传入天然气对象，则使用《GB/T 32151》中的定值粗略计算，如果传入天然气对象，则如下：
    natural_gas = NaturalGas(compositions={"CH4": 99, "N2": 1})
    碳排放量 = c_combustion_gas(flow_gas=10, natural_gas=natural_gas)


    :param natural_gas: 天然气对象
    :param flow_gas: Nm3/h
    :return: 碳排放量，t/h
    """

    flow_gas = flow_gas / 10000  # 万Nm3
    碳氧化率_天然气 = 0.99
    if natural_gas is None:
        ncv = 38931
        单位热值含碳量 = 15.3e-3
    elif isinstance(natural_gas, NaturalGas):
        natural_gas.set_pt(0.101325, 15)
        # ncv_m = natural_gas.get_ncv_mass(15)  # kJ/kg
        # ncv = ncv_m * natural_gas.get_density()  # kJ/kg/(Nm3/kg) = kJ/Nm3
        # c_ratio = natural_gas.get_ratio_carbon()  # 单位质量天然气中碳元素的质量, kg/kg
        # 单位热值含碳量 = c_ratio / ncv_m  # kg/kJ = t/MJ
        # 单位热值含碳量 = 单位热值含碳量 * 1000  # t/GJ
        c_ratio = natural_gas.get_ratio_carbon()
        density = natural_gas.get_density()
        return flow_gas * 10000 * density * c_ratio / 12 * 44 / 1000  # Nm3/h*(kg/Nm3)

    ncv = ncv / 100  # GJ/万Nm3

    天然气活动数据 = ncv * flow_gas  # t GJ
    排放因子 = 单位热值含碳量 * 碳氧化率_天然气 * 44 / 12
    E_燃烧 = 天然气活动数据 * 排放因子

    return E_燃烧, 排放因子


print(c_combustion_gas(10000))
natural_gas = NaturalGas({"CH4": 100, "N2": 0})
natural_gas.set_pt(0.101325, 15)
print(c_combustion_gas(10000, natural_gas))



import sys
import subprocess
import ctypes
import os
import pickle
import threading

# 由于模块的特殊性，不能引入自己编写的其他模块，否则会导致管理员身份运行失败

"""
以管理员方式运行cmd命令，使用方法如下：
import common.fileOperate as fo
data = {"popup": True, "command": "net start {}".format(serviceName)}
fo.writeAsPickle(temp_file, data)
iact.runAdmin()
即必须以临时文件传入命令及相关参数，否则无效
"""


def fprint(msg):
    # print(msg)
    msg = str(msg)
    with open("D:\\log_yangke.txt", "a") as f:
        f.write(msg + "\n\r")


def writeAsPickle(file: str, obj: object, opentype='wb'):
    with open(file, opentype) as f:
        pickle.dump(obj, f)


def readFromPickle(file: str, opentype='rb'):
    with open(file, opentype) as f:
        obj = pickle.load(f)
    return obj


def runAdmin(cmd=None, cwd=None, popup=True, python=None, charset="gbk", config=None, runas=False):
    """
    以管理员身份运行cmd命令，该方法总是会被执行两次，因为引入时会被执行一次

    popup为True时调试正常，运行不正常
    如果将popup设置为False，则不会弹出确认窗口，但是利用了windows的提权漏洞，可能会在以后的版本中被修复，具体参考https://yq.aliyun.com/articles/217616

    参考1：https://blog.csdn.net/MAOZEXIJR/article/details/88117388，该博客介绍的方法在调试时可以正常运行命令net start mysql，但直接运行时，mysql服务启动不起来，且运行不报错
    参考2：https://blog.csdn.net/qq_17550379/article/details/79006655?depth_1-utm_source=distribute.pc_relevant.none-task&utm_source=distribute.pc_relevant.none-task
        这里使用参考2中的方法，但稍作修改，主要是参考中使用的python版本是python2，测试发现，单独执行Interact.py可以，但是由第三方调用执行，则服务也启动不起来

    问题解决：发现第三方调用服务起不来的原因是，因为本文件以管理员身份运行时，相当于以__main__模式在运行该方法，因此，必须在
    if __name__=="__main__": 中再次调用runAdmin方法，第三方的调用只是运行了runAdmin()，儿runAdmin()中以管理员方式运行时重新
    调用了该__file__文件。

    :param cmd: 需要运行的命令
    :param cwd: 指定命令运行的目录
    :param popup: 是否弹出确认窗口，False时需要保证命令行中python命令对应的是项目对应的解释器
    :param config: 内置参数，无须传入值
    :param python: python解释器的路径，当本地有多个python解释器的时候，指定非默认解释器时使用，不指定则使用环境变量中的默认python解释器
    :param runas: 标记当前函数的调用者是不是__main__方法
    :return:
    """
    # 必须在第三方调用时将popup保存到另外一个独立文件中，这样当独立调用__file文件时，可以读入动态popup参数
    # temp_file = os.path.join(os.path.dirname(__file__), 'Interact_file')  # 这个文件路径和文件名必须和iact读取的一致
    config = "Interact_temp_file"
    runResult = config + "_runResult"
    cwd = cwd or os.getcwd()  # os.path.dirname(__file__)  # 这里赋值调用目录，不能是文件目录
    config = os.path.join(cwd, config)
    runResult = os.path.join(cwd, runResult)
    fprint("config_file=" + config + "; result_file=" + runResult)
    if cmd is not None:  # 说明是第三方模块调用，保存传入的参数信息
        # fprint("第一次：第三方模块调用")
        data = {"popup": popup, "command": cmd, "cwd": cwd, "charset": charset}
        writeAsPickle(config, data)
        fprint(readFromPickle(config))
        runAdmin(config=config, popup=popup, cwd=cwd, charset=charset)  # 这次调用仍不是管理员身份
        import time
        while not os.path.exists(runResult):  # 等待管理员身份运行结果
            time.sleep(0.1)
            fprint("wait for result...")
        result = readFromPickle(runResult)
        return result

    if os.path.exists(config):
        data = readFromPickle(config)
        popup = data['popup']
        cmd = data['command']
        cwd = data['cwd']
        charset = data['charset']

    if runas:  # 如果是以管理员身份从__main__模块运行的
        # 因为引入该模块时也会执行一次__main__模块，因此这里判断是不是引入时执行的，如果是引入执行的，则不存在config文件
        if os.path.exists(config):  # 如果不是引入该模块时执行的__main__方法
            # 从config文件读入参数
            fprint("第三次：自己以管理员方式运行文件时调用")
            fprint(config)
            data = readFromPickle(config)
            popup = data['popup']
            cmd = data['command']
            cwd = data['cwd']
            charset = data['charset']
            fprint("cmd={}, popup={}, cwd={}, charset={}".format(cmd, popup, cwd, charset))
        else:  # 如果是引入模块时调用的，则直接返回，不做操作
            fprint("第零次：引入时调用")
            fprint(os.path.abspath(config))
            return
    else:
        # 如果不是管理员身份运行的本文件，则此处不执行操作，由以下代码块切换为管理员身份重新运行，且runas设置为True
        fprint("第二次：自己递归调用自己")
        pass
    # 开始以管理员方法运行
    if popup:
        admin = is_admin()
        if admin:
            # 最好给该代码块加单线程锁
            sp = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            sp.wait()
            stderr = str(sp.stderr.read().decode(charset)).strip()
            stdout = str(sp.stdout.read().decode(charset)).strip()
            fprint(stdout)
            fprint(stderr)
            # 将输出写入文件runResult
            fprint("输出运行结果到文件：" + runResult)
            writeAsPickle(runResult, {"stdout": stdout, "stderr": stderr})
            return
        else:  # 使用管理员身份运行本文件
            # fprint("第二次调用完成，开始第三次调用，第三次调用没有调试信息，因为是独立进程")
            fprint("以管理员身份运行：" + __file__ + "工作目录：" + cwd)
            # 最好给下面代码加单线程锁
            pid = ctypes.windll.shell32.ShellExecuteW(None, "runas", sys.executable, __file__, cwd, 3)
            fprint(pid)
            return
    else:
        CMD = r"C:\Windows\System32\cmd.exe"
        FOD_HELPER = r'C:\Windows\System32\fodhelper.exe'
        PYTHON_CMD = python or "python"
        REG_PATH = 'Software\Classes\ms-settings\shell\open\command'
        DELEGATE_EXEC_REG_KEY = 'DelegateExecute'

        def create_reg_key(key, value):
            """
            Creates a reg key
            """
            try:
                import winreg
                winreg.CreateKey(winreg.HKEY_CURRENT_USER, REG_PATH)
                registry_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, REG_PATH, 0, winreg.KEY_WRITE)
                winreg.SetValueEx(registry_key, key, 0, winreg.REG_SZ, value)
                winreg.CloseKey(registry_key)
            except WindowsError:
                raise

        def bypass_uac(cmd):
            """
            Tries to bypass the UAC
            """
            try:
                create_reg_key(DELEGATE_EXEC_REG_KEY, '')
                create_reg_key(None, cmd)
            except WindowsError:
                raise

        def execute(cmd):
            if not is_admin():
                # 如果不是管理员身份，则以管理员身份重新运行当前文件，重新运行会进入if __name__ == __main__代码块
                fprint('[!] The script is NOT running with administrative privileges')
                fprint('[+] Trying to bypass the UAC')
                try:
                    current_dir = __file__
                    cmd1 = '{} /k "{}" "{}"'.format(CMD, PYTHON_CMD, current_dir)
                    bypass_uac(cmd1)
                    os.system(FOD_HELPER)
                    sys.exit(0)
                except WindowsError:
                    sys.exit(1)
            else:
                # 这里添加我们需要管理员权限的代码
                fprint('[+] The script is running with administrative privileges!')
                sp = subprocess.Popen(
                    cmd,
                    shell=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE
                )
                sp.wait()
                stderr = str(sp.stderr.read().decode(charset)).strip()
                stdout = str(sp.stdout.read().decode(charset)).strip()
                fprint(stdout)
                fprint(stderr)
                # 将输出写入文件runResult
                writeAsPickle(runResult, {"stdout": stdout, "stderr": stderr})

        execute(cmd)


def is_admin():
    """判断当前运行环境是否具有管理员权限"""
    try:
        o = ctypes.windll.shell32.IsUserAnAdmin()
        if o == 1:
            result = True
        elif o == 0:
            result = False
    except:
        return False
    return result


fprint("__name__=" + __name__)
mutex = threading.Lock()
times = 0

if __name__ == "__main__":
    mutex.acquire()
    fprint("进入__main__模块")
    runAdmin(runas=True)
    mutex.release()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2019/8/20
# @Author  : github.com/guofei9987


import numpy as np
from .base import SkoBase
from yangke.sko.tools import func_transformer
from yangke.common.config import logger
from abc import ABCMeta, abstractmethod
from .operators import crossover, mutation, ranking, selection


class GeneticAlgorithmBase(SkoBase, metaclass=ABCMeta):
    def __init__(self, func, n_dim,
                 size_pop=50, max_iter=200, prob_mut=0.001,
                 constraint_eq=tuple(), constraint_ueq=tuple()):
        self.func = func_transformer(func)
        self.size_pop = size_pop  # size of population
        self.max_iter = max_iter
        self.prob_mut = prob_mut  # probability of mutation
        self.n_dim = n_dim

        # constraint:
        self.has_constraint = len(constraint_eq) > 0 or len(constraint_ueq) > 0
        self.constraint_eq = list(constraint_eq)  # a list of equal functions with ceq[i] = 0
        self.constraint_ueq = list(constraint_ueq)  # a list of unequal constraint functions with c[i] <= 0

        self.Chrom = None
        self.X = None  # shape = (size_pop, n_dim)
        self.Y_raw = None  # shape = (size_pop,) , value is f(x)
        self.Y = None  # shape = (size_pop,) , value is f(x) + penalty for constraint
        self.FitV = None  # shape = (size_pop,)

        # self.FitV_history = []
        self.generation_best_X = []
        self.generation_best_Y = []

        self.all_history_Y = []
        self.all_history_FitV = []

        self.best_x, self.best_y = None, None

    @abstractmethod
    def chrom2x(self, Chrom):
        pass

    def x2y(self):
        self.Y_raw = self.func(self.X)
        if not self.has_constraint:
            self.Y = self.Y_raw
        else:
            # constraint
            penalty_eq = np.array([np.sum(np.abs([c_i(x) for c_i in self.constraint_eq])) for x in self.X])
            penalty_ueq = np.array([np.sum(np.abs([max(0, c_i(x)) for c_i in self.constraint_ueq])) for x in self.X])
            self.Y = self.Y_raw + 1e5 * penalty_eq + 1e5 * penalty_ueq
        return self.Y

    @abstractmethod
    def ranking(self):
        pass

    @abstractmethod
    def selection(self):
        pass

    @abstractmethod
    def crossover(self):
        pass

    @abstractmethod
    def mutation(self):
        pass

    def run(self, max_iter=None):
        self.max_iter = max_iter or self.max_iter
        for i in range(self.max_iter):
            logger.debug(f"计算第{i}代种群...")
            self.X = self.chrom2x(self.Chrom)
            self.Y = self.x2y()
            self.ranking()
            self.selection()
            self.crossover()
            self.mutation()

            # record the best ones
            generation_best_index = self.FitV.argmax()
            self.generation_best_X.append(self.X[generation_best_index, :])
            self.generation_best_Y.append(self.Y[generation_best_index])
            self.all_history_Y.append(self.Y)
            self.all_history_FitV.append(self.FitV)

        global_best_index = np.array(self.generation_best_Y).argmin()
        self.best_x = self.generation_best_X[global_best_index]
        self.best_y = self.func(np.array([self.best_x]))
        return self.best_x, self.best_y

    fit = run


class GA(GeneticAlgorithmBase):
    """
    genetic algorithm

    Parameters
    ----------------
    func : function
        The func you want to do optimal
    n_dim : int
        number of variables of func
    lb : array_like
        The lower bound of every variables of func
    ub : array_like
        The upper bound of every variables of func
    constraint_eq : tuple
        equal constraint
    constraint_ueq : tuple
        unequal constraint
    precision : array_like
        The precision of every variables of func
    size_pop : int
        Size of population
    max_iter : int
        Max of iter
    prob_mut : float between 0 and 1
        Probability of mutation
    Attributes
    ----------------------
    Lind : array_like
         The num of genes of every variable of func（segments）
    generation_best_X : array_like. Size is max_iter.
        Best X of every generation
    generation_best_ranking : array_like. Size if max_iter.
        Best ranking of every generation
    Examples
    -------------
    https://github.com/guofei9987/scikit-opt/blob/master/examples/demo_ga.py

    约束条件使用：如果有以下约束条件
    x0*x1>=1
    x0*x1<=5
    x1+x2=1
    则：
    constraint_eq = [
        lambda x: 1-x[1]-x[2]
    ]
    constraint_ueq = [
        lambda x: 1-x[0]*x[1],
        lambda x: x[0]*x[1]-5
    ]  # 不等式约束的值要小于0
    """

    def __init__(self, func, n_dim,
                 size_pop=50, max_iter=200,
                 prob_mut=0.001,
                 lb=-1, ub=1,
                 constraint_eq=tuple(), constraint_ueq=tuple(),
                 precision=1e-7):
        super().__init__(func, n_dim, size_pop, max_iter, prob_mut, constraint_eq, constraint_ueq)

        self.lb, self.ub = np.array(lb) * np.ones(self.n_dim), np.array(ub) * np.ones(self.n_dim)
        self.precision = np.array(precision) * np.ones(self.n_dim)  # works when precision is int, float, list or array

        # Lind是基因长度
        Lind_raw = np.log2((self.ub - self.lb) / self.precision + 1)
        self.Lind = np.ceil(Lind_raw).astype(int)

        # if precision is integer:
        # if Lind_raw is integer, which means the number of all possible value is 2**n, no need to modify
        # if Lind_raw is decimal, we need ub_extend to make the number equal to 2**n,
        self.int_mode_ = (self.precision % 1 == 0) & (Lind_raw % 1 != 0)
        self.int_mode = np.any(self.int_mode_)
        if self.int_mode:
            self.ub_extend = np.where(self.int_mode_
                                      , self.lb + (np.exp2(self.Lind) - 1) * self.precision
                                      , self.ub)

        self.len_chrom = sum(self.Lind)

        self.crtbp()

    def crtbp(self):
        # create the population
        self.Chrom = np.random.randint(low=0, high=2, size=(self.size_pop, self.len_chrom))
        return self.Chrom

    def gray2rv(self, gray_code):
        # Gray Code to real value: one piece of a whole chromosome
        # input is a 2-dimensional numpy array of 0 and 1.
        # output is a 1-dimensional numpy array which convert every row of input into a real number.
        _, len_gray_code = gray_code.shape
        # b = gray_code.cumsum(axis=1) % 2  # 不知道为什么官方代码中添加这一句进行基因转换，感觉没啥作用且增加了程序复杂度
        # 经检索资料，这里是将二进制编码转换为格雷码
        b = gray_code
        mask = np.logspace(start=1, stop=len_gray_code, base=0.5, num=len_gray_code)
        return (b * mask).sum(axis=1) / mask.sum()

    def chrom2x(self, Chrom):
        cumsum_len_segment = self.Lind.cumsum()
        X = np.zeros(shape=(self.size_pop, self.n_dim))
        for i, j in enumerate(cumsum_len_segment):
            if i == 0:
                Chrom_temp = Chrom[:, :cumsum_len_segment[0]]
            else:
                Chrom_temp = Chrom[:, cumsum_len_segment[i - 1]:cumsum_len_segment[i]]
            X[:, i] = self.gray2rv(Chrom_temp)

        if self.int_mode:
            X = self.lb + (self.ub_extend - self.lb) * X
            X = np.where(X > self.ub, self.ub, X)
            # the ub may not obey precision, which is ok.
            # for example, if precision=2, lb=0, ub=5, then x can be 5
        else:
            X = self.lb + (self.ub - self.lb) * X
        # c_ = self.x2chrom_single(X[0])
        return X

    def set_chrom(self, index, x):
        """
        设置GA的初代种群中，第index个个体为制定的个体x，x为实际的数值列表，不是基因编码序列

        :param index:
        :param x:
        :return:
        """
        code = self.x2chrom_single(x)
        self.Chrom[index] = code

    def x2chrom_single(self, x):
        """
        将一组输入参数转换为基因表达，即二进制数组，例如优化变量为3个，则x=[24, 100, 32]
        self.x2chrom_single(x) 将x转换为一个基因序列。
        本方法本质上是chrom2x的逆反运算，但chrom2x一次转换种群数量个个体，而本方法一次只转换一个个体。

        :param x:
        :return:
        """
        if self.int_mode:
            x = (x - self.lb) / (self.ub_extend - self.lb)
            logger.debug(f"暂不支持整数类型的变量回算!")
            ...
        else:
            x = (x - self.lb) / (self.ub - self.lb)

        chroms = []
        # 将数值转换为二进制
        for i, data in enumerate(x):  # 将x逐个转换为二进制数组
            mask = np.logspace(start=1, stop=self.Lind[i], base=0.5, num=self.Lind[i])
            _ = mask.sum() * data
            chrom = []
            for b in list(mask):
                if _ >= b:
                    chrom.append(1)
                    _ = _ - b
                else:
                    chrom.append(0)
            chroms.extend(chrom)
        return chroms

    ranking = ranking.ranking
    selection = selection.selection_tournament_faster
    crossover = crossover.crossover_2point_bit
    mutation = mutation.mutation

    def to(self, device):
        '''
        use pytorch to get parallel performance
        '''
        try:
            import torch
            from .operators_gpu import crossover_gpu, mutation_gpu, selection_gpu, ranking_gpu
        except:
            print('pytorch is needed')
            return self

        self.device = device
        self.Chrom = torch.tensor(self.Chrom, device=device, dtype=torch.int8)

        def chrom2x(self, Chrom):
            '''
            We do not intend to make all operators as tensor,
            because objective function is probably not for pytorch
            '''
            Chrom = Chrom.cpu().numpy()
            cumsum_len_segment = self.Lind.cumsum()
            X = np.zeros(shape=(self.size_pop, self.n_dim))
            for i, j in enumerate(cumsum_len_segment):
                if i == 0:
                    Chrom_temp = Chrom[:, :cumsum_len_segment[0]]
                else:
                    Chrom_temp = Chrom[:, cumsum_len_segment[i - 1]:cumsum_len_segment[i]]
                X[:, i] = self.gray2rv(Chrom_temp)

            if self.int_mode:
                X = self.lb + (self.ub_extend - self.lb) * X
                X = np.where(X > self.ub, self.ub, X)
            else:
                X = self.lb + (self.ub - self.lb) * X
            return X

        self.register('mutation', mutation_gpu.mutation). \
            register('crossover', crossover_gpu.crossover_2point_bit). \
            register('chrom2x', chrom2x)

        return self


class GA_TSP(GeneticAlgorithmBase):
    """
    Do genetic algorithm to solve the TSP (Travelling Salesman Problem)
    Parameters
    ----------------
    func : function
        The func you want to do optimal.
        It inputs a candidate solution(a routine), and return the costs of the routine.
    size_pop : int
        Size of population
    max_iter : int
        Max of iter
    prob_mut : float between 0 and 1
        Probability of mutation
    Attributes
    ----------------------
    Lind : array_like
         The num of genes corresponding to every variable of func（segments）
    generation_best_X : array_like. Size is max_iter.
        Best X of every generation
    generation_best_ranking : array_like. Size if max_iter.
        Best ranking of every generation
    Examples
    -------------
    Firstly, your data (the distance matrix). Here I generate the data randomly as a demo:
    ```py
    num_points = 8
    points_coordinate = np.random.rand(num_points, 2)  # generate coordinate of points
    distance_matrix = spatial.distance.cdist(points_coordinate, points_coordinate, metric='euclidean')
    print('distance_matrix is: \n', distance_matrix)
    def cal_total_distance(routine):
        num_points, = routine.shape
        return sum([distance_matrix[routine[i % num_points], routine[(i + 1) % num_points]] for i in range(num_points)])
    ```
    Do GA
    ```py
    from sko.GA import GA_TSP
    ga_tsp = GA_TSP(func=cal_total_distance, n_dim=8, pop=50, max_iter=200, Pm=0.001)
    best_points, best_distance = ga_tsp.run()
    ```
    """

    def __init__(self, func, n_dim, size_pop=50, max_iter=200, prob_mut=0.001):
        super().__init__(func, n_dim, size_pop=size_pop, max_iter=max_iter, prob_mut=prob_mut)
        self.has_constraint = False
        self.len_chrom = self.n_dim
        self.crtbp()

    def crtbp(self):
        # create the population
        tmp = np.random.rand(self.size_pop, self.len_chrom)
        self.Chrom = tmp.argsort(axis=1)
        return self.Chrom

    def chrom2x(self, Chrom):
        return Chrom

    ranking = ranking.ranking
    selection = selection.selection_tournament_faster
    crossover = crossover.crossover_pmx
    mutation = mutation.mutation_reverse

    def run(self, max_iter=None):
        self.max_iter = max_iter or self.max_iter
        for i in range(self.max_iter):
            Chrom_old = self.Chrom.copy()
            self.X = self.chrom2x(self.Chrom)
            self.Y = self.x2y()
            self.ranking()
            self.selection()
            self.crossover()
            self.mutation()

            # put parent and offspring together and select the best size_pop number of population
            self.Chrom = np.concatenate([Chrom_old, self.Chrom], axis=0)
            self.X = self.chrom2x(self.Chrom)
            self.Y = self.x2y()
            self.ranking()
            selected_idx = np.argsort(self.Y)[:self.size_pop]
            self.Chrom = self.Chrom[selected_idx, :]

            # record the best ones
            generation_best_index = self.FitV.argmax()
            self.generation_best_X.append(self.X[generation_best_index, :].copy())
            self.generation_best_Y.append(self.Y[generation_best_index])
            self.all_history_Y.append(self.Y.copy())
            self.all_history_FitV.append(self.FitV.copy())

        global_best_index = np.array(self.generation_best_Y).argmin()
        self.best_x = self.generation_best_X[global_best_index]
        self.best_y = self.func(np.array([self.best_x]))
        return self.best_x, self.best_y



# 设置模型路径
import numpy as np
import pandas as pd

from yangke.common.config import logger
from yangke.ebsilon.ebsilon import EbsUnits, EbsApp, EbsModel, EbsObject

model_path = r"C:\Users\YangKe\Desktop\ebs\冷端优化-100%试验工况.ebs"
# data_path = r"C:\Users\杨可\Documents\WPS Cloud Files\217145378\6科研项目\2021\华能重庆两江燃机发电有限责任公司\重庆两江批量计算.xlsx"
data_path = r"C:\Users\YangKe\Documents\WPS Cloud Files\217145378\漫游文档\其他设备\重庆两江批量计算1.xlsx"


class Condition:
    def __init__(self, name):
        self.name = name
        self.data = []

    def append_parameters(self, component, variable, value, unit=None, profile=None):
        self.data.append({"component": component, "variable": variable, "value": value, "unit": unit})


def init_ebsilon():
    """
    初始化ebsilon
    :return:
    """
    ebsilon = EbsApp()
    ebsilon.describe()
    return ebsilon


def read_data_from_xlsx(filename: str):
    """
    读取xlsx文件中的数据，并组装为Conditions字典
    :param filename:
    :return:
    """
    sheet_data: pd.DataFrame = pd.read_excel(filename)
    sheet_data.dropna(subset=["变量名", "组件名"], inplace=True)  # 删除变量名和组件名为空的行
    sheet_data.dropna(axis=1, how="all")
    conditions = {}
    for name in sheet_data.columns[4:]:
        conditions.update({name: Condition(name)})
    for i, row in sheet_data.iterrows():
        # i：行号
        # row=Series(('参数', '压力') ('组件名', 'fgh') ('变量名', 'MEASM') ('单位', 'MPa') ('工况1', 3.8) ('工况2', 3.2)...)
        # row中前三项定义了本行是模型中哪个组件的哪个参数以及参数单位
        # row中从第四项开始是各个工况的数值
        component_name = row["组件名"]
        variable_name = row["变量名"]
        unit = row["单位"]

        for col in row[4:].iteritems():  # row中从第四项开始是各个工况的值
            conditions.get(col[0]).append_parameters(component_name, variable_name, col[1], unit)
    return conditions


def solve(ebsilon, condition):
    model_path_test = f"{model_path[:-4]}_{condition.NAME}.ebs"

    # 打开模型文件，并激活design的profile
    model: EbsModel = ebsilon.open(model_path)
    model.activate_profile("design")  # 激活当前操作的Profile，后续操作均在激活的Profile上进行

    # 设置组件参数
    for para in condition.data:
        comp = para.get("component")
        var = para.get("variable")
        value = para.get("value")
        unit = para.get("unit")
        model.set_value(comp, var, value, unit, save_flag=False)

    # 获取模型中的参数值
    # p_fgh_gas_in = model.get_value("p_fgh_in_gas", "MEASM", EbsUnits.MPa)
    # power1 = model.get_value("Generator_1","QREAL", EbsUnits.MW)

    model.save_as(model_path_test)  # 另存一份模型，后续在新模型上更改
    ebsilon.show_window()

    ebs_calc_result = model.simulate_new()
    logger.info(ebs_calc_result)

    power1 = power1 = model.get_value("Generator_1", "QREAL", EbsUnits.MW)
    logger.info(f"燃机发电机功率为{power1} MW")

    model.save()
    logger.info(f"工况{condition.NAME}计算完成！")


data: dict = read_data_from_xlsx(data_path)
ebs = init_ebsilon()
for key, condition in data.items():
    solve(ebs, condition)
